import unittest

from biogeme.catalog import Catalog, generic_alt_specific_catalogs
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta
from biogeme.segmentation import DiscreteSegmentationTuple


class TestGenericAltSpecificCatalog(unittest.TestCase):

    def setUp(self):
        # Create dummy Beta parameters
        self.beta1 = Beta('beta1', 1, None, None, 0)
        self.beta2 = Beta('beta2', 2, None, None, 0)
        self.alternatives = ('Car', 'Train')
        self.segmentations = (
            DiscreteSegmentationTuple(variable='income', mapping={0: 'low', 1: 'high'}),
        )

    def test_error_if_too_few_alternatives(self):
        with self.assertRaises(BiogemeError) as cm:
            generic_alt_specific_catalogs(
                'test',
                [self.beta1],
                ('OnlyOneAlt',),
            )
        self.assertIn('requires at least 2 alternatives', str(cm.exception))

    def test_error_if_not_list(self):
        with self.assertRaises(BiogemeError) as cm:
            generic_alt_specific_catalogs(
                'test',
                'not_a_list',
                self.alternatives,
            )
        self.assertIn('must be a list', str(cm.exception))

    def test_error_if_not_beta(self):
        with self.assertRaises(BiogemeError) as cm:
            generic_alt_specific_catalogs(
                'test',
                [self.beta1, 42],
                self.alternatives,
            )
        self.assertIn('not Beta expressions', str(cm.exception))

    def test_return_structure_without_segmentation(self):
        catalogs = generic_alt_specific_catalogs(
            'test', [self.beta1, self.beta2], self.alternatives
        )
        self.assertEqual(len(catalogs), 2)
        for item in catalogs:
            self.assertIsInstance(item, dict)
            self.assertEqual(set(item.keys()), set(self.alternatives))
            for alt, cat in item.items():
                self.assertIsInstance(cat, Catalog)
                self.assertIn('generic', [expr.name for expr in cat])
                self.assertIn('altspec', [expr.name for expr in cat])

    def test_return_structure_with_segmentation(self):
        catalogs = generic_alt_specific_catalogs(
            'test',
            [self.beta1],
            self.alternatives,
            potential_segmentations=self.segmentations,
        )
        self.assertEqual(len(catalogs), 1)
        self.assertIsInstance(catalogs[0], dict)
        for alt in self.alternatives:
            names = [expr.name for expr in catalogs[0][alt]]
            self.assertIsInstance(catalogs[0][alt], Catalog)
            self.assertIn('generic', names)
            self.assertIn('altspec', names)


if __name__ == '__main__':
    unittest.main()



"""
Implementation of the main Biogeme class

:author: Michel Bierlaire
:date: Tue Mar 26 16:45:15 2019

It combines the database and the model specification.
"""

from __future__ import annotations

import difflib
import logging
import warnings
from datetime import datetime

import numpy as np
import pandas as pd
from tqdm import tqdm

from biogeme.biogeme_logging import suppress_logs
from biogeme.calculator import (
    CallableExpression,
    CompiledFormulaEvaluator,
    MultiRowEvaluator,
    function_from_compiled_formula,
)
from biogeme.calculator.simple_formula import evaluate_simple_expression
from biogeme.catalog import (
    CentralController,
    Configuration,
    SelectedConfigurationsIterator,
)
from biogeme.constants import LOG_LIKE, WEIGHT
from biogeme.database import Database
from biogeme.default_parameters import ParameterValue
from biogeme.deprecated import deprecated_parameters
from biogeme.dict_of_formulas import check_validity, get_expression
from biogeme.draws import RandomNumberGeneratorTuple
from biogeme.exceptions import BiogemeError, ValueOutOfRange
from biogeme.expressions import (
    Expression,
    ExpressionOrNumeric,
    IntegrateNormal,
    MultipleSum,
    log,
)
from biogeme.filenames import get_new_file_name
from biogeme.function_output import (
    FunctionOutput,
)
from biogeme.likelihood import AlgorithmResults, model_estimation
from biogeme.likelihood.bootstrap import bootstrap
from biogeme.model_elements import ModelElements
from biogeme.optimization import OptimizationAlgorithm, algorithms
from biogeme.parameters import (
    DEFAULT_FILE_NAME as DEFAULT_PARAMETER_FILE_NAME,
    Parameters,
)
from biogeme.results_processing import (
    RawEstimationResults,
    generate_html_file,
)
from biogeme.results_processing.estimation_results import EstimationResults
from biogeme.second_derivatives import SecondDerivativesMode
from biogeme.tools import (
    CheckDerivativesResults,
    ModelNames,
    check_derivatives,
    safe_serialize_array,
)
from biogeme.tools.files import files_of_type
from biogeme.validation import ValidationResult, cross_validate_model

DEFAULT_MODEL_NAME = 'biogeme_model_default_name'
logger = logging.getLogger(__name__)


class BIOGEME:
    """Main class that combines the database and the model
        specification.

    It works in two modes: estimation and simulation.

    The following attributes are imported from the parameter file.


    """

    # Type hints for dynamically injected attributes
    __annotations__ = {
        'seed': int,
        'tolerance': float,
        'steptol': float,
        'max_iterations': int,
        'number_of_threads': int,
        'enlarging_factor': float,
        'initial_radius': float,
        'infeasible_cg': bool,
        'second_derivatives': int,
        'dogleg': bool,
        'bootstrap_samples': int,
        'save_iterations': bool,
        'generate_html': bool,
        'generate_yaml': bool,
        'optimization_algorithm': str,
        'maximum_number_catalog_expressions': int,
        'max_number_parameters_to_report': int,
    }

    def __init__(
        self,
        database: Database,
        formulas: Expression | dict[str, Expression],
        random_number_generators: dict[str:RandomNumberGeneratorTuple] | None = None,
        user_notes: str | None = None,
        parameters: str | Parameters | None = None,
        **kwargs,
    ):
        """Constructor

        :param database: choice data.
        :param formulas: expression or dictionary of expressions that
             define the model specification.  The concept is that each
             expression is applied to each entry of the database. The
             keys of the dictionary allow to provide a name to each
             formula.  In the estimation mode, two formulas are
             needed, with the keys 'loglike' and 'weight'. If only one
             formula is provided, it is associated with the label
             'loglike'. If no formula is labeled 'weight', the weight
             of each piece of data is supposed to be 1.0. In the
             simulation mode, the labels of each formula are used as
             labels of the resulting database.

        :param random_number_generators: user defined random number generators.
        :param user_notes: these notes will be included in the report file.
        :param parameters: name of the .toml file where the parameters are read, or Parameters object containing
        :raise BiogemeError: an audit of the formulas is performed.
           If a formula has issues, an error is detected and an
           exception is raised.

        """
        if isinstance(parameters, Parameters):
            logger.info('Biogeme parameters provided by the user.')
            self.biogeme_parameters: Parameters = parameters
        else:
            self.biogeme_parameters: Parameters = Parameters()
            if isinstance(parameters, str):
                self.biogeme_parameters.read_file(parameters)
            else:
                if parameters is None:
                    self.biogeme_parameters.read_file(DEFAULT_PARAMETER_FILE_NAME)
                else:
                    error_msg = (
                        f'Argument "parameters" is of wrong type: {type(parameters)}'
                    )
                    raise AttributeError(error_msg)

        database.reset_indices()
        self.parameter_file: str = self.biogeme_parameters.file_name
        self.html_filename: str | None = None
        self.yaml_filename: str | None = None

        # We allow the values of the parameters to be set with arguments
        self.biogeme_parameters.set_several_parameters(dict_of_parameters=kwargs)

        for name in self.biogeme_parameters.parameter_names:
            self._define_property(name)

        self.maximum_number_of_observations_per_individual: int | None = None
        self.database = database

        self.log_like_name: str = LOG_LIKE
        """ Keywords used for the name of the loglikelihood formula.
        Default: 'log_like'"""

        self.log_like_valid_names: list[str] = [LOG_LIKE, 'loglike']

        self.weight_name = WEIGHT
        """Keyword used for the name of the weight formula. Default: 'weight'
        """

        self.weight_valid_names = [WEIGHT, 'weights']

        self.model_name = DEFAULT_MODEL_NAME
        """Name of the model. Default: 'biogemeModelDefaultName'
        """

        self.second_derivatives_mode = SecondDerivativesMode(
            self.calculating_second_derivatives
        )

        if self.seed != 0:
            np.random.seed(self.seed)

        self.short_names: ModelNames | None = None

        self.formulas = self._normalize_formulas(formulas)

        self.user_notes = user_notes  #: User notes

        self.init_loglikelihood = None  #: Init value of the likelihood function

        self.null_loglikelihood = None  #: Log likelihood of the null model

        self.best_iteration = None  #: Store the best iteration found so far.

        self._model_elements = ModelElements(
            expressions=self.formulas,
            database=self.database,
            number_of_draws=self.biogeme_parameters.get_value(name='number_of_draws'),
            user_defined_draws=random_number_generators,
        )
        self._function_evaluator = (
            CompiledFormulaEvaluator(
                model_elements=self._model_elements,
                second_derivatives_mode=self.second_derivatives_mode,
                numerically_safe=self.numerically_safe,
            )
            if self.contains_log_likelihood()
            else None
        )

    @classmethod
    def from_configuration_and_controller(
        cls,
        config_id: str,
        central_controller: CentralController,
        database: Database,
        user_notes: str | None = None,
        parameters: str | Parameters | None = None,
        **kwargs,
    ) -> BIOGEME:
        """Obtain the Biogeme object corresponding to the
        configuration of a multiple expression

        :param config_id: identifier of the configuration

        :param central_controller: central controller for the multiple expression containing all the catalogs.

        :param database: database to be passed to the Biogeme object

        :param user_notes: these notes will be included in the report file.

        :param parameters: object with the parameters

        """
        if central_controller.all_configurations:
            # We verify that the configuration is valid
            the_set = central_controller.all_configurations_ids
            if not the_set:
                error_msg = "No configuration found in the expression"
                raise BiogemeError(error_msg)
            if config_id not in the_set:
                close_matches = difflib.get_close_matches(config_id, the_set)
                if close_matches:
                    error_msg = (
                        f"Unknown configuration: [{config_id}]. "
                        f"Did you mean [{close_matches[0]}]?"
                    )
                else:
                    error_msg = f"Unknown configuration: {config_id}."
                raise BiogemeError(error_msg)
        the_configuration = Configuration.from_string(config_id)
        central_controller.set_configuration(the_configuration)
        if user_notes is None:
            user_notes = the_configuration.get_html()
        flat_expression = central_controller.expression.deep_flat_copy()
        return cls(
            database=database,
            formulas=flat_expression,
            user_notes=user_notes,
            parameters=parameters,
            **kwargs,
        )

    @classmethod
    def from_configuration(
        cls,
        config_id: str,
        multiple_expression: Expression,
        database: Database,
        user_notes: str | None = None,
        parameters: str | Parameters | None = None,
        **kwargs,
    ) -> BIOGEME:
        """Obtain the Biogeme object corresponding to the
        configuration of a multiple expression
        :param config_id: identifier of the configuration

        :param multiple_expression: multiple expression containing the catalog.

        :param database: database to be passed to the Biogeme object

        :param user_notes: these notes will be included in the report file.

        :param parameters: object with the parameters

        """
        central_controller = CentralController(expression=multiple_expression)
        return cls.from_configuration_and_controller(
            config_id=config_id,
            central_controller=central_controller,
            database=database,
            user_notes=user_notes,
            parameters=parameters,
            **kwargs,
        )

    def _define_property(self, name: str):
        def getter(obj):
            return obj.biogeme_parameters.get_value(name=name)

        def setter(obj, value):
            error_msg = (
                f"Direct assignment to '{name}' is not allowed. Please set the value in the "
                f"constructor or in the .toml file."
            )
            raise BiogemeError(error_msg)

        prop = property(getter, setter)
        setattr(self.__class__, name, prop)

    @property
    def generate_pickle(self) -> bool:
        warnings.warn(
            "'generate_pickle' is deprecated. Use 'generate_yaml' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return False

    @generate_pickle.setter
    def generate_pickle(self, value: bool) -> None:
        warnings.warn(
            "'generate_pickle' is deprecated. Use 'generate_yaml' instead. This statement is ignored",
            DeprecationWarning,
            stacklevel=2,
        )

    @property
    def modelName(self) -> str:
        return self.model_name

    @modelName.setter
    def modelName(self, value: str) -> None:
        warnings.warn(
            "'modelName' is deprecated. Please use 'model_name' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.model_name = value

    @property
    def expressions_registry(self):
        return self._model_elements.expressions_registry

    @property
    def log_like(self) -> Expression | None:
        return get_expression(
            dict_of_formulas=self.formulas, valid_keywords=self.log_like_valid_names
        )

    def contains_log_likelihood(self) -> bool:
        return self.log_like is not None

    @property
    def weight(self) -> Expression | None:
        return get_expression(
            dict_of_formulas=self.formulas, valid_keywords=self.weight_valid_names
        )

    @property
    def model_elements(self) -> ModelElements | None:
        return self._model_elements

    @property
    def function_evaluator(self) -> CompiledFormulaEvaluator:
        return self._function_evaluator

    def _normalize_formulas(
        self, formulas: Expression | dict[str, Expression]
    ) -> dict[str, Expression]:
        """
        Normalize user input to a dictionary of expressions.
        Raises BiogemeError if input is invalid.

        :param formulas: a single expression or a dictionary of named expressions.
        :return: a dictionary mapping names to expressions.
        """
        if isinstance(formulas, Expression):
            return {self.log_like_name: formulas}
        if isinstance(formulas, dict):
            check_validity(formulas)
            return formulas
        raise BiogemeError(
            f'Invalid type for formulas: {type(formulas)}. Expected Expression or dict.'
        )

    def is_model_complex(self) -> bool:
        """Check if the model is potentially complex to estimate"""
        if self.log_like.requires_draws():
            return True
        if self.log_like.embed_expression(IntegrateNormal):
            return True

        return False

    @property
    def loglike(self) -> Expression:
        """For backward compatibility"""
        return self.log_like

    @property
    def function_parameters(self) -> dict[str, ParameterValue]:
        """Prepare the parameters for the function"""
        return {
            "tolerance": self.biogeme_parameters.get_value('tolerance'),
            "steptol": self.biogeme_parameters.get_value('steptol'),
        }

    @property
    def algo_parameters(self) -> dict[str, ParameterValue]:
        """Prepare the parameters for the optimization algorithm."""
        common_bounds_params = {
            'infeasibleConjugateGradient': self.biogeme_parameters.get_value(
                'infeasible_cg'
            ),
            'radius': self.biogeme_parameters.get_value('initial_radius'),
            'enlargingFactor': self.biogeme_parameters.get_value('enlarging_factor'),
            'maxiter': self.biogeme_parameters.get_value('max_iterations'),
            'max_number_parameters_to_report': self.biogeme_parameters.get_value(
                'max_number_parameters_to_report'
            ),
        }

        if self.optimization_algorithm == 'automatic':
            if (
                self.is_model_complex()
                or self.second_derivatives_mode == SecondDerivativesMode.NEVER
            ):
                logger.info(
                    'As the model is rather complex, we cancel the calculation of second derivatives. '
                    'If you want to control the parameters, change the algorithm from "automatic" '
                    'to "simple_bounds" in the TOML file.'
                )
                algo_parameters = common_bounds_params | {
                    "proportionAnalyticalHessian": 0,
                }
            else:
                logger.info(
                    'As the model is not too complex, we activate the calculation of second derivatives. '
                    'To change this behavior, modify the algorithm to "simple_bounds" in the TOML file.'
                )
                algo_parameters = common_bounds_params | {
                    "proportionAnalyticalHessian": 1,
                }
            return algo_parameters

        if self.optimization_algorithm == "simple_bounds":
            algo_parameters = common_bounds_params | {
                "proportionAnalyticalHessian": self.biogeme_parameters.get_value(
                    'second_derivatives'
                ),
            }
            if (
                self.second_derivatives_mode == SecondDerivativesMode.FINITE_DIFFERENCES
                and self.biogeme_parameters.get_value('second_derivatives') > 0
            ):
                warning = (
                    f'The proportion of the analytical hessian is not zero, and the second derivatives are approximated by '
                    f'finite difference. It may not be the desired configuration.'
                )
                logger.warning(warning)
            if (
                self.second_derivatives_mode == SecondDerivativesMode.NEVER
                and self.biogeme_parameters.get_value('second_derivatives') > 0
            ):
                error_msg = (
                    f'The proportion of the analytical hessian is not zero, and the second derivatives cannot be '
                    f'evaluated. The parameters "calculating_second_derivatives" and "second_derivatives" are inconsistent.'
                )
            return algo_parameters

        if self.optimization_algorithm in {
            "simple_bounds_newton",
            "simple_bounds_BFGS",
        }:
            return common_bounds_params

        if self.optimization_algorithm in {"TR-newton", "TR-BFGS"}:
            algo_parameters = common_bounds_params | {
                "dogleg": self.biogeme_parameters.get_value('dogleg'),
                "radius": self.biogeme_parameters.get_value('initial_radius'),
                "maxiter": self.biogeme_parameters.get_value('max_iterations'),
            }
            return algo_parameters

        if self.optimization_algorithm in {"LS-newton", "LS-BFGS"}:
            return common_bounds_params

        return common_bounds_params

    @property
    def optimization_parameters(self) -> dict[str, ParameterValue]:
        return self.function_parameters | self.algo_parameters

    def _save_iterations_file_name(self) -> str:
        """
        :return: The name of the file where the iterations are saved.
        :rtype: str
        """
        return f"__{self.model_name}.iter"

    @property
    def free_betas_names(self) -> list[str]:
        """Returns the names of the parameters that must be estimated

        :return: list of names of the parameters
        :rtype: list(str)
        """
        return self.expressions_registry.free_betas_names

    def number_unknown_parameters(self) -> int:
        """Returns the number of parameters that must be estimated

        :return: number of parameters
        :rtype: int
        """
        return self.expressions_registry.number_of_free_betas

    def calculate_null_loglikelihood(self, avail: dict[int, ExpressionOrNumeric]):
        """Calculate the log likelihood of the null model that predicts equal
        probability for each alternative

        :param avail: list of expressions to evaluate the availability
                      conditions for each alternative. If 1 is provided, it is always available.
        :type avail: list of :class:`biogeme.expressions.Expression`

        :return: value of the log likelihood
        :rtype: float

        """
        expression = -log(MultipleSum(avail))
        result = evaluate_simple_expression(
            expression=expression, database=self.database, numerically_safe=False
        )
        self.null_loglikelihood = result
        return self.null_loglikelihood

    def calculate_init_likelihood(self) -> float:
        """Calculate the value of the log likelihood function

        The default values of the parameters are used.

        :return: value of the log likelihood.
        :rtype: float.
        """
        # Value of the loglikelihood for the default values of the parameters.

        result = self.function_evaluator.evaluate(
            the_betas=self.expressions_registry.free_betas_init_values,
            gradient=False,
            hessian=False,
            bhhh=False,
        )
        return result.function

    def _get_likelihood_function(
        self,
    ) -> CallableExpression:
        return function_from_compiled_formula(
            the_compiled_function=self.function_evaluator,
            the_betas=self.expressions_registry.free_betas_init_values,
        )

    def report_array(self, array: np.ndarray, with_names: bool = True) -> str:
        """Reports the entries of the array up to the maximum number

        :param array: array to report
        :type array: numpy.array

        :param with_names: if True, the names of the parameters are included
        :type with_names: bool

        :return: string reporting the values
        :rtype: str
        """
        length = min(
            array.size,
            self.biogeme_parameters.get_value('max_number_parameters_to_report'),
        )
        if with_names:
            names = self.expressions_registry.free_betas_names
            report = ", ".join(
                [
                    f"{name}={value:.2g}"
                    for name, value in zip(names[:length], array[:length])
                ]
            )
            return report
        report = ", ".join([f"{value:.2g}" for value in array[:length]])
        return report

    def _load_saved_iteration(self) -> dict[str, float]:
        """Reads the values of the parameters from a text file where each line
        has the form name_of_beta = value_of_beta, and use these values in all
        formulas.

        """
        filename = self._save_iterations_file_name()
        betas = {}
        try:
            with open(filename, encoding="utf-8") as fp:
                for line in fp:
                    ell = line.split("=")
                    betas[ell[0].strip()] = float(ell[1])
            self.change_init_values(betas)
            logger.info(f"Parameter values restored from {filename}")
            return betas
        except OSError:
            logger.info(f"Cannot read file {filename}. Statement is ignored.")
            return {}

    def set_random_init_values(self, default_bound: float = 100.0) -> None:
        """Modifies the initial values of the parameters in all formulas,
        using randomly generated values. The value is drawn from a
        uniform distribution on the interval defined by the
        bounds.

        :param default_bound: If the upper bound is missing, it is
            replaced by this value. If the lower bound is missing, it is
            replaced by the opposite of this value. Default: 100.
        :type default_bound: float
        """
        random_betas = {
            beta.name: np.random.uniform(
                low=-default_bound if beta.lower_bound is None else beta.lower_bound,
                high=default_bound if beta.upper_bound is None else beta.upper_bound,
            )
            for beta in self.expressions_registry.free_betas
        }
        self.change_init_values(random_betas)

    def change_init_values(self, betas: dict[str, float]) -> None:
        """Modifies the initial values of the parameters in all formula

        :param betas: dictionary where the keys are the names of the
                      parameters, and the values are the new value for
                      the parameters.
        :type betas: dict(string:float)
        """
        for expression in self.formulas.values():
            expression.change_init_values(betas)
        for beta in self.expressions_registry.free_betas:
            value = betas.get(beta.name)
            if value is not None:
                beta.init_value = value

    def _load_saved_estimates(self) -> EstimationResults:
        yaml_files = files_of_type('yaml', name=self.model_name)
        yaml_files.sort()
        if yaml_files:
            yaml_to_read = yaml_files[-1]
            if len(yaml_files) > 1:
                warning_msg = (
                    f"Several files .yaml are available for "
                    f"this model: {yaml_files}. "
                    f"The file {yaml_to_read} "
                    f"is used to load the results."
                )
                logger.warning(warning_msg)
            results = EstimationResults.from_yaml_file(filename=yaml_to_read)
            logger.warning(
                f'Estimation results read from {yaml_to_read}. '
                f'There is no guarantee that they correspond '
                f'to the specified model.'
            )
            return results
        raise BiogemeError(f'No yaml file has been found for model {self.model_name}')

    def _bootstrap(self, estimated_parameters: dict[str, float]) -> list[np.ndarray]:
        number_of_bootstrap_samples = self.biogeme_parameters.get_value(
            'bootstrap_samples'
        )
        if number_of_bootstrap_samples == 0:
            return []

        logger.info(
            f"Re-estimate the model {self.biogeme_parameters.get_value('bootstrap_samples')} "
            f"times for bootstrapping"
        )
        with suppress_logs(level=logging.WARNING):
            # For some reason, the self.optimization_parameters cannot be used as such in the function call below.
            # I have no clue why.
            opt_parameters = self.optimization_parameters
            bootstrap_results: list[AlgorithmResults] = bootstrap(
                number_of_bootstrap_samples=number_of_bootstrap_samples,
                the_algorithm=self._algorithm_configuration(),
                modeling_elements=self.model_elements,
                parameters=opt_parameters,
                starting_values=estimated_parameters,
                second_derivatives_mode=self.biogeme_parameters.get_value(
                    name='calculating_second_derivatives'
                ),
                numerically_safe=self.numerically_safe,
                number_of_jobs=self.biogeme_parameters.get_value(name='number_of_jobs'),
            )

        return [result.solution for result in bootstrap_results]

    def retrieve_saved_estimates(self) -> EstimationResults | None:
        """
        Attempt to retrieve previously saved estimation results from a YAML file.

        :return:
            An EstimationResults object if a saved result is found. If no file is found or loading fails,
            None is returned and a warning is logged.

        :raises BiogemeError:
            Raised internally by _load_saved_estimates if loading fails, and is caught to log a warning instead.
        """
        try:
            return self._load_saved_estimates()
        except BiogemeError as e:
            logger.warning(e)
            return None

    def estimate(
        self,
        starting_values: dict[str, float] | None = None,
        recycle: bool = False,
        run_bootstrap: bool = False,
        **kwargs,
    ) -> EstimationResults:
        """Estimate the parameters of the model(s).

        :return: object containing the estimation results.
        :rtype: biogeme.bioResults

        Example::

            # Create an instance of biogeme
            biogeme  = bio.BIOGEME(database, logprob)

            # Gives a name to the model
            biogeme.modelName = 'mymodel'

            # Estimate the parameters
            results = biogeme.estimate()

        :raises BiogemeError: if no expression has been provided for the
            likelihood

        """

        if kwargs.get("bootstrap") is not None:
            error_msg = (
                'Parameter "bootstrap" is deprecated. In order to perform '
                "bootstrapping, bootstrap_samples=100 to a positive number in the biogeme.toml file ["
                "e.g. bootstrap_samples=100]."
            )
            raise BiogemeError(error_msg)
        if kwargs.get("algorithm") is not None:
            error_msg = (
                'The parameter "algorithm" is deprecated. Instead, define the '
                'parameter "optimization_algorithm" in section "[Estimation]" '
                "of the TOML parameter file"
            )
            raise BiogemeError(error_msg)

        if kwargs.get("algo_parameters") is not None:
            error_msg = (
                'The parameter "algo_parameters" is deprecated. Instead, define the '
                'parameters "max_iterations" and "tolerance" in section '
                '"[SimpleBounds]" '
                "of the TOML parameter file"
            )
            raise BiogemeError(error_msg)

        if kwargs.get("algoParameters") is not None:
            error_msg = (
                'The parameter "algoParameters" is deprecated. Instead, define the '
                'parameters "max_iterations" and "tolerance" in section '
                '"[SimpleBounds]" '
                "of the TOML parameter file"
            )
            raise BiogemeError(error_msg)

        if kwargs:
            unexpected = ', '.join(kwargs.keys())
            raise BiogemeError(
                f'Ignoring unexpected arguments passed to estimate(): {unexpected}. '
                f'This method does not accept any parameters. Use biogeme.toml or the BIOGEME object to set parameters.'
            )

        if self.log_like is None:
            raise BiogemeError("No log likelihood function has been specified")

        if recycle:
            try:
                return self._load_saved_estimates()
            except BiogemeError as e:
                logger.warning(e)

        if self.model_name == DEFAULT_MODEL_NAME:
            logger.warning(
                f'You have not defined a name for the model. '
                f'The output files are named from the model name. '
                f'The default is [{DEFAULT_MODEL_NAME}]'
            )

        if self.expressions_registry.number_of_free_betas == 0:
            raise BiogemeError(
                f"There is no parameter to estimate"
                f" in the formula: {self.log_like}."
            )

        save_iteration_file_name = None
        saved_starting_values = {}
        if self.biogeme_parameters.get_value('save_iterations'):
            logger.info(
                f"*** Initial values of the parameters are "
                f"obtained from the file {self._save_iterations_file_name()}"
            )
            save_iteration_file_name = self._save_iterations_file_name()
            saved_starting_values = self._load_saved_iteration()

        if starting_values is None:
            starting_values = saved_starting_values.copy()
        else:
            for key, value in saved_starting_values.items():
                starting_values.setdefault(key, value)
        logger.info(f'Starting values for the algorithm: {starting_values}')
        init_log_likelihood = self.calculate_init_likelihood()
        logger.debug(f'Init log likelihood: {init_log_likelihood}')

        # For some reason, the self.optimization_parameters cannot be used as such in the function call below.
        # I have no clue why.
        opt_parameters = self.optimization_parameters

        algorithm_results: AlgorithmResults = model_estimation(
            the_algorithm=self._algorithm_configuration(),
            function_evaluator=self.function_evaluator,
            parameters=opt_parameters,
            some_starting_values=starting_values,
            save_iterations_filename=save_iteration_file_name,
        )

        optimal_betas = self.expressions_registry.get_named_betas_values(
            algorithm_results.solution
        )

        calculate_hessian = self.second_derivatives_mode != SecondDerivativesMode.NEVER
        f_g_h_b: FunctionOutput = self.function_evaluator.evaluate(
            the_betas=optimal_betas,
            gradient=True,
            hessian=calculate_hessian,
            bhhh=True,
        )

        if run_bootstrap:
            start_time = datetime.now()
            bootstrap_results = self._bootstrap(estimated_parameters=optimal_betas)
            # Time needed to generate the bootstrap results
            bootstrap_time = datetime.now() - start_time
        else:
            bootstrap_results = []
            bootstrap_time = None

        clean_optimization_messages = {}
        for key, value in algorithm_results.optimization_messages.items():
            if isinstance(value, np.floating):
                clean_optimization_messages[key] = float(value)
            elif isinstance(value, np.ndarray):
                clean_optimization_messages[key] = value.tolist()
            else:
                clean_optimization_messages[key] = value

        null_log_likelihood = (
            float(self.null_loglikelihood)
            if self.null_loglikelihood is not None
            else None
        )
        the_hessian = (
            None if f_g_h_b.hessian is None else safe_serialize_array(f_g_h_b.hessian)
        )
        raw_estimation_results = RawEstimationResults(
            model_name=self.model_name,
            user_notes=self.user_notes,
            beta_names=self.expressions_registry.free_betas_names,
            beta_values=algorithm_results.solution.tolist(),
            lower_bounds=[bound[0] for bound in self.expressions_registry.bounds],
            upper_bounds=[bound[1] for bound in self.expressions_registry.bounds],
            gradient=safe_serialize_array(f_g_h_b.gradient),
            hessian=the_hessian,
            bhhh=safe_serialize_array(f_g_h_b.bhhh),
            null_log_likelihood=null_log_likelihood,
            initial_log_likelihood=init_log_likelihood,
            final_log_likelihood=f_g_h_b.function,
            data_name=self.model_elements.database.name,
            sample_size=self.model_elements.sample_size,
            number_of_observations=self.model_elements.number_of_observations,
            monte_carlo=self.expressions_registry.requires_draws,
            number_of_draws=int(self.model_elements.draws_management.number_of_draws),
            types_of_draws=self.model_elements.draws_management.draw_types,
            number_of_excluded_data=self.model_elements.database.number_of_excluded_data,
            draws_processing_time=self.model_elements.draws_management.processing_time,
            optimization_messages=clean_optimization_messages,
            convergence=algorithm_results.convergence,
            bootstrap=[one_result.tolist() for one_result in bootstrap_results],
            bootstrap_time=bootstrap_time,
        )
        estimation_results = EstimationResults(
            raw_estimation_results=raw_estimation_results
        )
        estimated_betas = estimation_results.get_beta_values()
        for f in self.formulas.values():
            f.change_init_values(estimated_betas)

        if not estimation_results.algorithm_has_converged:
            logger.warning(
                'It seems that the optimization algorithm did not converge. '
                'Therefore, the results may not correspond to the maximum '
                'likelihood estimator. Check the specification of the model, '
                'or the criteria for convergence of the algorithm.'
            )
        if self.biogeme_parameters.get_value('generate_html'):
            self.html_filename = get_new_file_name(self.model_name, 'html')
            generate_html_file(
                filename=self.html_filename, estimation_results=estimation_results
            )
        if self.biogeme_parameters.get_value('generate_yaml'):
            self.yaml_filename = get_new_file_name(self.model_name, 'yaml')
            estimation_results.dump_yaml_file(filename=self.yaml_filename)
        return estimation_results

    def quick_estimate(self) -> EstimationResults:
        """| Estimate the parameters of the model. Same as estimate, where any
             extra calculation is skipped (init loglikelihood,
             t-statistics, etc.)

        :return: object containing the estimation results.
        :rtype: biogeme.results.bioResults

        Example::

            # Create an instance of biogeme
            biogeme  = bio.BIOGEME(database, logprob)

            # Gives a name to the model
            biogeme.modelName = 'mymodel'

            # Estimate the parameters
            results = biogeme.quickEstimate()

        :raises BiogemeError: if no expression has been provided for the
            likelihood

        """
        # For some reason, the self.optimization_parameters cannot be used as such in the function call below.
        # I have no clue why.
        save_iteration_file_name = None
        starting_values = {}
        if self.biogeme_parameters.get_value('save_iterations'):
            logger.info(
                f"*** Initial values of the parameters are "
                f"obtained from the file {self._save_iterations_file_name()}"
            )
            save_iteration_file_name = self._save_iterations_file_name()
            starting_values = self._load_saved_iteration()

        # For some reason, the self.optimization_parameters cannot be used as such in the function call below.
        # I have no clue why.
        opt_parameters = self.optimization_parameters

        algorithm_results: AlgorithmResults = model_estimation(
            the_algorithm=self._algorithm_configuration(),
            function_evaluator=self.function_evaluator,
            parameters=opt_parameters,
            some_starting_values=starting_values,
            save_iterations_filename=save_iteration_file_name,
        )

        optimal_betas = self.expressions_registry.get_named_betas_values(
            algorithm_results.solution
        )
        f_g_h_b: FunctionOutput = self.function_evaluator.evaluate(
            the_betas=optimal_betas,
            gradient=False,
            hessian=False,
            bhhh=False,
        )

        clean_optimization_messages = {}
        for key, value in algorithm_results.optimization_messages.items():
            if isinstance(value, np.floating):
                clean_optimization_messages[key] = float(value)
            elif isinstance(value, np.ndarray):
                clean_optimization_messages[key] = value.tolist()
            else:
                clean_optimization_messages[key] = value

        null_log_likelihood = (
            float(self.null_loglikelihood)
            if self.null_loglikelihood is not None
            else None
        )
        raw_estimation_results = RawEstimationResults(
            model_name=self.model_name,
            user_notes=self.user_notes,
            beta_names=self.expressions_registry.free_betas_names,
            beta_values=algorithm_results.solution.tolist(),
            lower_bounds=[bound[0] for bound in self.expressions_registry.bounds],
            upper_bounds=[bound[1] for bound in self.expressions_registry.bounds],
            gradient=[],
            hessian=[[]],
            bhhh=[[]],
            null_log_likelihood=null_log_likelihood,
            initial_log_likelihood=None,
            final_log_likelihood=f_g_h_b.function,
            data_name=self.model_elements.database.name,
            sample_size=self.model_elements.sample_size,
            number_of_observations=self.model_elements.number_of_observations,
            monte_carlo=self.expressions_registry.requires_draws,
            number_of_draws=int(self.model_elements.draws_management.number_of_draws),
            types_of_draws=self.model_elements.draws_management.draw_types,
            number_of_excluded_data=self.model_elements.database.number_of_excluded_data,
            draws_processing_time=self.model_elements.draws_management.processing_time,
            optimization_messages=clean_optimization_messages,
            convergence=algorithm_results.convergence,
            bootstrap=[],
            bootstrap_time=None,
        )
        return EstimationResults(raw_estimation_results=raw_estimation_results)

    def estimate_catalog(
        self,
        selected_configurations: set[Configuration] = None,
        quick_estimate: bool = False,
        recycle: bool = False,
        run_bootstrap: bool = False,
    ) -> dict[str, EstimationResults]:
        """Estimate all or selected versions of a model with Catalog's,
        corresponding to multiple specifications.

        :param selected_configurations: set of configurations. If
            None, all configurations are considered.
        :param quick_estimate: if True, the final statistics are not calculated.
        :param recycle: if True, the results are read from the pickle
            file, if it exists. If False, the estimation is performed.
        :param run_bootstrap: if True, bootstrapping is applied.
        :return: object containing the estimation results associated
            with the name of each specification, as well as a
            description of each configuration

        """
        if self.short_names is None:
            self.short_names = ModelNames(prefix=self.modelName)

        if self.log_like is None:
            raise BiogemeError("No log likelihood function has been specified")

        central_controller = CentralController(
            expression=self.log_like,
            maximum_number_of_configurations=self.biogeme_parameters.get_value(
                'maximum_number_catalog_expressions'
            ),
        )
        if selected_configurations is None:
            number_of_specifications = central_controller.number_of_configurations()
            logger.info(f'Estimating {number_of_specifications} models.')

            if (
                number_of_specifications is None
                or number_of_specifications > self.maximum_number_catalog_expressions
            ):
                error_msg = (
                    f"There are too many [{number_of_specifications}] different "
                    f"specifications for the log likelihood function. This is "
                    f"above the maximum number: "
                    f"{self.maximum_number_catalog_expressions}. Simplify "
                    f"the specification, change the value of the parameter "
                    f"maximum_number_catalog_expressions, or consider using "
                    f'the AssistedSpecification object in the "biogeme.assisted" '
                    f"module."
                )
                raise ValueOutOfRange(error_msg)

            the_iterator = SelectedConfigurationsIterator(
                the_central_controller=central_controller
            )
        else:
            the_iterator = SelectedConfigurationsIterator(
                the_central_controller=central_controller,
                selected_configurations=selected_configurations,
            )
        configurations = {}
        for config in the_iterator:
            config_id = config.get_string_id()
            b = self.__class__.from_configuration(
                config_id=config_id,
                multiple_expression=self.log_like,
                database=self.database,
                user_notes=self.user_notes,
                parameters=self.biogeme_parameters,
            )
            b.model_name = self.short_names(config_id)
            results = b.retrieve_saved_estimates() if recycle else None
            if results is None:
                results = (
                    b.quick_estimate()
                    if quick_estimate
                    else b.estimate(recycle=recycle, run_bootstrap=run_bootstrap)
                )

            configurations[config_id] = results

        return configurations

    def validate(
        self,
        estimation_results: EstimationResults,
        slices: int,
        groups: str | None = None,
    ) -> list[ValidationResult]:
        """
        Perform out-of-sample validation of the model.

        The validation procedure operates by dividing the dataset into a number of slices.
        For each slice:
          - The slice is used as the validation set.
          - The remaining data forms the estimation set.
          - The model is re-estimated on the estimation set.
          - The model is applied to the validation set to compute the log likelihood.

        :param estimation_results: Estimation results obtained from the full dataset.
        :param slices: Number of data splits to create for cross-validation.
        :param groups: Optional column name used to group data entries (e.g., panel data). If provided, splitting preserves groups.

        :return: List of validation results, one for each data slice.
        :raises BiogemeError: If the dataset is structured as panel data and incompatible with validation.
        """
        # For some reason, the self.optimization_parameters cannot be used as such in the function call below.
        # I have no clue why.
        parameters = self.optimization_parameters
        parameters['calculating_second_derivatives'] = (
            self.biogeme_parameters.get_value(name='calculating_second_derivatives')
        )
        return cross_validate_model(
            the_algorithm=self._algorithm_configuration(),
            modeling_elements=self.model_elements,
            parameters=parameters,
            starting_values=estimation_results.get_beta_values(),
            slices=slices,
            numerically_safe=self.numerically_safe,
            groups=groups,
        )

    def _algorithm_configuration(self) -> OptimizationAlgorithm:
        if self.biogeme_parameters.get_value('optimization_algorithm') == "automatic":
            self.biogeme_parameters.set_value('optimization_algorithm', 'simple_bounds')
        return algorithms.get(self.optimization_algorithm)

    def simulate(self, the_beta_values: dict[str, float] | None) -> pd.DataFrame:
        """Evaluate all simulation formulas on each row of the database using the specified parameter values.

        :param the_beta_values: Dictionary mapping parameter names to values.
                                If None, an exception is raised. Use results.get_beta_values()
                                after estimation or provide explicit values.

        :return: A pandas DataFrame where each row corresponds to an observation in the database,
                 and each column corresponds to a simulation formula.

        :raises BiogemeError: If the_beta_values is None or if the number of parameters is incorrect.
        """
        if the_beta_values is None:
            current_beta_values = self.expressions_registry.free_betas
            err = (
                f'Contrarily to previous versions of Biogeme, '
                f'the values of Beta must '
                f'now be explicitly mentioned. If they have been estimated, they can be obtained from '
                f'results.get_beta_values(). If not, use the default values: {current_beta_values}'
            )
            raise BiogemeError(err)

        if not isinstance(the_beta_values, dict):
            error_msg = f'the_beta_values must be a dict, and not an object of type {type(the_beta_values)}'
            raise BiogemeError(error_msg)

        the_evaluator: MultiRowEvaluator = MultiRowEvaluator(
            model_elements=self.model_elements, numerically_safe=self.numerically_safe
        )

        results = the_evaluator.evaluate(the_beta_values)
        return results

    def confidence_intervals(
        self, beta_values: list[dict[str, float]], interval_size: float = 0.9
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Calculate confidence intervals on the simulated quantities


        :param beta_values: array of parameters values to be used in
               the calculations. Typically, it is a sample drawn from
               a distribution.
        :type beta_values: list(dict(str: float))

        :param interval_size: size of the reported confidence interval,
                    in percentage. If it is denoted by s, the interval
                    is calculated for the quantiles (1-s)/2 and
                    (1+s)/2. The default (0.9) corresponds to
                    quantiles for the confidence interval [0.05, 0.95].
        :type interval_size: float

        :return: two pandas data frames 'left' and 'right' with the
            same dimensions. Each row corresponds to a row in the
            database, and each column to a formula. 'left' contains the
            left value of the confidence interval, and 'right' the right
            value

            Example::

                # Read the estimation results from a file
                results = EstimationEResults.from_yaml_file(filename = 'my_model.yaml')
                # Retrieve the names of the betas parameters that have been
                # estimated
                betas = biogeme.freeBetaNames

                # Draw 100 realization of the distribution of the estimators
                b = results.getBetasForSensitivityAnalysis(betas, size = 100)

                # Simulate the formulas using the nominal values
                simulatedValues = biogeme.simulate(beta_values)

                # Calculate the confidence intervals for each formula
                left, right = biogeme.confidenceIntervals(b, 0.9)

        :rtype: tuple of two Pandas dataframes.

        """
        list_of_results = []
        for b in tqdm(beta_values):
            r = self.simulate(b)
            list_of_results += [r]
        all_results = pd.concat(list_of_results)
        r = (1.0 - interval_size) / 2.0
        left = all_results.groupby(level=0).quantile(r)
        right = all_results.groupby(level=0).quantile(1.0 - r)
        return left, right

    def __str__(self) -> str:
        r = f"{self.model_name}: database [{self.model_elements.database.name}]"
        r += str(self.formulas)
        return r

    @deprecated_parameters({'beta': None})
    def check_derivatives(self, verbose: bool = False) -> CheckDerivativesResults:
        """Verifies the implementation of the derivatives.

        It compares the analytical version with the finite differences
        approximation.

        :param verbose: if True, the comparisons are reported. Default: False.

        :return: f, g, h, gdiff, hdiff where

            - f is the value of the function,
            - g is the analytical gradient,
            - h is the analytical hessian,
            - gdiff is the difference between the analytical and the
              finite differences gradient,
            - hdiff is the difference between the analytical and the
              finite differences hessian,

        """

        starting_values = (
            self.model_elements.expressions_registry.complete_dict_of_free_beta_values(
                the_betas={}
            )
        )

        the_log_likelihood: CallableExpression = function_from_compiled_formula(
            the_compiled_function=self.function_evaluator,
            the_betas=starting_values,
        )

        def the_function(x: np.ndarray) -> FunctionOutput:
            """Wrapper function to use tools.checkDerivatives"""
            the_function_output: FunctionOutput = the_log_likelihood(
                x, gradient=True, hessian=True, bhhh=False
            )
            return the_function_output

        return check_derivatives(
            the_function,
            np.asarray(
                self.model_elements.expressions_registry.list_of_free_betas_init_values
            ),
            self.model_elements.expressions_registry.free_betas_names,
            verbose,
        )



""" Implements various models.

:author: Michel Bierlaire
:date: Fri Mar 29 17:13:14 2019
"""

import logging
from typing import Callable
import biogeme.distributions as dist
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, Beta, NormalCdf

logger = logging.getLogger(__name__)


def ordered_likelihood(
    continuous_value: Expression,
    list_of_discrete_values: list[int],
    tau_parameter: Beta,
    cdf: Callable[[Expression], Expression],
) -> dict[int, Expression]:
    """Ordered model that maps a continuous quantity with a list of
        discrete intervals (often logit or probit)

    Example: discrete values = [1, 2, 3, 4]

    We define thresholds tau_1_2, tau_2_3 and tau_3_4.
    In order to impose that the threshold are sorted, we actually define
        tau_1_2 = tau_parameter
        tau_2_3 = tau_1_2 + diff2
        tau_3_4 = tau_2_3 + diff3

    The probability that the discrete value is 2, say, is the
    probability that the continuous value lies between tau_1_2 and
    tau_2_3, where the probability distribution is logistic.

    :param continuous_value: continuous quantity to mapping

    :param list_of_discrete_values: discrete values

    :param tau_parameter: parameter for the first threshold

    :param cdf: function calculating the CDF of the random variable

    :return: dict where the keys are the discrete values and the
        values are the corresponding probability.
    """
    if not isinstance(tau_parameter, Beta):
        error_msg = (
            f'tau_parameter must be a Beta expression, and not a {type(tau_parameter)}.'
        )
        raise BiogemeError(error_msg)

    if len(list_of_discrete_values) == 2:
        the_proba = {
            list_of_discrete_values[0]: 1 - cdf(continuous_value - tau_parameter),
            list_of_discrete_values[1]: cdf(continuous_value - tau_parameter),
        }

        return the_proba

    diffs = {
        current_item: Beta(
            f'{tau_parameter.name}_diff_{current_item}',
            1,
            0,
            None,
            0,
        )
        for current_item in list_of_discrete_values[1:-1]
    }

    # First term
    the_proba = {list_of_discrete_values[0]: 1 - cdf(continuous_value - tau_parameter)}

    # Intermediate terms
    tau = tau_parameter
    for item in list_of_discrete_values[1:-1]:
        next_tau = tau + diffs[item]
        the_proba[item] = cdf(continuous_value - tau) - cdf(continuous_value - next_tau)
        tau = next_tau

    # Last term
    the_proba[list_of_discrete_values[-1]] = cdf(continuous_value - tau)

    return the_proba


def ordered_logit(
    continuous_value: Expression,
    list_of_discrete_values: list[int],
    tau_parameter: Beta,
) -> dict[int, Expression]:
    """Ordered logit model that maps a continuous quantity with a
        list of discrete intervals

    Example: discrete values = [1, 2, 3, 4]

    We define thresholds tau_1_2, tau_2_3 and tau_3_4.
    In order to impose that the threshold are sorted, we actually define
        tau_1_2 = tau_parameter
        tau_2_3 = tau_1_2 + diff2
        tau_3_4 = tau_2_3 + diff3

    The probability that the discrete value is 2, say, is the
    probability that the continuous value lies between tau_1_2 and
    tau_2_3, where the probability distribution is logistic.

    :param continuous_value: continuous quantity to mapping

    :param list_of_discrete_values: discrete values

    :param tau_parameter: parameter for the first threshold

    """
    return ordered_likelihood(
        continuous_value=continuous_value,
        list_of_discrete_values=list_of_discrete_values,
        tau_parameter=tau_parameter,
        cdf=dist.logisticcdf,
    )


def ordered_probit(
    continuous_value: Expression,
    list_of_discrete_values: list[int],
    tau_parameter: Beta,
) -> dict[int, Expression]:
    """Ordered probit model that maps a continuous quantity with a
        list of discrete intervals

    Example: discrete values = [1, 2, 3, 4]

    We define thresholds tau_1_2, tau_2_3 and tau_3_4.
    In order to impose that the threshold are sorted, we actually define
        tau_1_2 = tau_parameter
        tau_2_3 = tau_1_2 + diff2
        tau_3_4 = tau_2_3 + diff3

    The probability that the discrete value is 2, say, is the
    probability that the continuous value lies between tau_1_2 and
    tau_2_3, where the probability distribution is normal.

    :param continuous_value: continuous quantity to mapping

    :param list_of_discrete_values: discrete values

    :param tau_parameter: parameter for the first threshold

    """
    return ordered_likelihood(
        continuous_value=continuous_value,
        list_of_discrete_values=list_of_discrete_values,
        tau_parameter=tau_parameter,
        cdf=NormalCdf,
    )



"""

biogeme.segmentation
====================

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Thu Dec  7 15:18:45 2023
"""
from biogeme.version import getText
import biogeme.segmentation as seg
from biogeme.expressions import Variable, Beta

# %%
# Version of Biogeme.
print(getText())

# %%
socio_eco_1 = Variable('socio_eco_1')

# %%
segmentation_1 = seg.DiscreteSegmentationTuple(
    variable=socio_eco_1,
    mapping={
        1: 'high',
        2: 'medium',
        3: 'low',
    },
)

# %%
socio_eco_2 = Variable('socio_eco_2')

# %%
segmentation_2 = seg.DiscreteSegmentationTuple(
    variable=socio_eco_2,
    mapping={
        1: 'level_1',
        2: 'level_2',
        3: 'level_3',
        4: 'level_4',
    },
)

# %%
beta_x = Beta('beta_x', 0, None, None, 0)

# %%
segmented_parameter = seg.Segmentation(beta_x, (segmentation_1,))

# %%
# The expressions for the segmented parameter is obtained as follows:
segmented_parameter.segmented_beta()

# %%
# The corresponding Python code can be obtained as follows.
print(segmented_parameter.segmented_code())

# %%
# The code of the original beta is also available:
print(segmented_parameter.beta_code())

# %%
# Same with the second segmentation
segmented_parameter = seg.Segmentation(beta_x, (segmentation_2,))

# %%
segmented_parameter.segmented_beta()

# %%
print(segmented_parameter.segmented_code())

# %%
# The two segmentations can be combined together
segmented_parameter = seg.Segmentation(
    beta_x,
    (
        segmentation_1,
        segmentation_2,
    ),
)

# %%
segmented_parameter.segmented_beta()

# %%
print(segmented_parameter.segmented_code())



"""
Test the "deprecated" decorators"

:author: Michel Bierlaire
:date: Mon May 13 09:35:05 2024

"""

import warnings

from biogeme.deprecated import deprecated, deprecated_parameters

import unittest
from unittest.mock import patch


class TestValidateParamsDecorator(unittest.TestCase):
    def test_valid_parameters(self):
        @deprecated_parameters(obsolete_params={})
        def func(a, b):
            return a + b

        self.assertEqual(func(a=1, b=2), 3)

    def test_valid_parameters_default(self):
        @deprecated_parameters(obsolete_params={})
        def func(a, b=2):
            return a + b

        self.assertEqual(func(a=1), 3)

    def test_deprecated_parameters(self):
        @deprecated_parameters(obsolete_params={'old_name': 'new_name'})
        def func(new_name):
            return new_name * 2

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            result = func(old_name=5)
            self.assertEqual(result, 10)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))

    def test_deprecated_parameters_mixed(self):
        @deprecated_parameters(
            obsolete_params={'old_name_a': 'new_name_a', 'old_name_b': 'new_name_b'}
        )
        def func(new_name_a, new_name_b):
            return new_name_a * new_name_b

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            result = func(old_name_a=5, new_name_b=2)
            self.assertEqual(result, 10)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            result = func(old_name_a=5, old_name_b=2)
            self.assertEqual(result, 10)
            self.assertEqual(len(w), 2)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))

    def test_deprecated_parameters_None(self):
        @deprecated_parameters(obsolete_params={'old_name': None})
        def func(x):
            return 2 * x

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            result = func(x=3, old_name=5)
            self.assertEqual(result, 6)

            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))

        with self.assertRaises(TypeError):
            result = func(old_name=5)


if __name__ == '__main__':
    unittest.main()



"""Specification of a multiple expressions used for testing purposes

"""

from biogeme.expressions import Beta
from biogeme import models
import biogeme.segmentation as seg
from biogeme.catalog import segmentation_catalogs

from swissmetro_data import (
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    MALE,
)

ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

gender_segmentation = seg.DiscreteSegmentationTuple(
    variable=MALE,
    mapping={
        0: 'female',
        1: 'male',
    },
)


ASC_CAR_catalog, ASC_TRAIN_catalog = segmentation_catalogs(
    generic_name='ASC',
    beta_parameters=[ASC_CAR, ASC_TRAIN],
    potential_segmentations=(gender_segmentation,),
    maximum_number=1,
)

# Definition of the utility functions with linear cost
V1 = ASC_TRAIN_catalog + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR_catalog + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

logprob = models.loglogit(V, av, CHOICE)



"""

Investigation of several choice models
======================================

Investigate several choice models:

    - logit
    - nested logit with two nests: public and private transportation
    - nested logit with two nests existing and future modes

for a total of 3 specifications.
See `Bierlaire and Ortelli (2023)
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_.

:author: Michel Bierlaire, EPFL
:date: Fri Jul 14 09:47:21 2023

"""
import biogeme.biogeme as bio
import biogeme.biogeme_logging as blog
from biogeme import models
from biogeme.expressions import Beta
from biogeme.catalog import Catalog
from biogeme.nests import OneNestForNestedLogit, NestsForNestedLogit
from biogeme.results import compile_estimation_results, pareto_optimal

# %%
# See :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)

# %%
# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the logit model. This is the contribution of each
# observation to the log likelihood function.
logprob_logit = models.loglogit(V, av, CHOICE)

# %%
# Nested logit model: nest with existing alternatives.
mu_existing = Beta('mu_existing', 1, 1, 10, 0)
existing = OneNestForNestedLogit(
    nest_param=mu_existing, list_of_alternatives=[1, 3], name='Existing'
)

nests_existing = NestsForNestedLogit(choice_set=list(V), tuple_of_nests=(existing,))
logprob_nested_existing = models.lognested(V, av, nests_existing, CHOICE)

# %%
# Nested logit model: nest with public transportation alternatives.
mu_public = Beta('mu_public', 1, 1, 10, 0)
public = OneNestForNestedLogit(
    nest_param=mu_public, list_of_alternatives=[1, 2], name='Public'
)

nests_public = NestsForNestedLogit(choice_set=list(V), tuple_of_nests=(public,))
logprob_nested_public = models.lognested(V, av, nests_public, CHOICE)

# %%
# Catalog.
model_catalog = Catalog.from_dict(
    catalog_name='model_catalog',
    dict_of_expressions={
        'logit': logprob_logit,
        'nested existing': logprob_nested_existing,
        'nested public': logprob_nested_public,
    },
)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, model_catalog)
the_biogeme.modelName = 'b01model'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# %%
# Estimate the parameters.
dict_of_results = the_biogeme.estimate_catalog()

# %%
# Number of estimated models.
print(f'A total of {len(dict_of_results)} models have been estimated')

# %%
# All estimation results
compiled_results, specs = compile_estimation_results(
    dict_of_results, use_short_names=True
)
# %%
compiled_results

# %%
# Glossary
for short_name, spec in specs.items():
    print(f'{short_name}\t{spec}')

# %%
# Estimation results of the Pareto optimal models.
pareto_results = pareto_optimal(dict_of_results)
compiled_pareto_results, pareto_specs = compile_estimation_results(
    pareto_results, use_short_names=True
)

# %%
compiled_pareto_results

# %%
# Glossary.
for short_name, spec in pareto_specs.items():
    print(f'{short_name}\t{spec}')



"""
Stores and process the estimation results

Michel Bierlaire
Mon Sep 30 15:39:24 2024
"""

from __future__ import annotations

import warnings
from typing import Any

import numpy as np
import pandas as pd
from biogeme.deprecated import deprecated_parameters
from biogeme.exceptions import BiogemeError
from biogeme.filenames import get_new_file_name
from biogeme.tools import likelihood_ratio
from biogeme.tools.ellipse import Ellipse
from numpy.linalg import LinAlgError, eigh, inv, norm, pinv
from scipy.stats import chi2, norm as normal_distribution
from yaml.constructor import ConstructorError

from .raw_estimation_results import (
    RawEstimationResults,
    deserialize_from_yaml,
    serialize_to_yaml,
)
from .recycle_pickle import read_pickle_biogeme
from .variance_covariance import EstimateVarianceCovariance


def calc_p_value(t: float) -> float:
    """Calculates the p value of a parameter from its t-statistic.

    The formula is

    .. math:: 2(1-\\Phi(|t|))

    where :math:`\\Phi(\\cdot)` is the CDF of a normal distribution.

    :param t: t-statistics
    :type t: float

    :return: p-value
    :rtype: float
    """
    p_value = 2.0 * (1.0 - normal_distribution.cdf(abs(t)))
    return p_value


def calculates_correlation_matrix(covariance: np.ndarray) -> np.ndarray:
    """Calculates the correlation matrix."""
    d = np.diag(covariance)
    if (d > 0).all():
        diagonal = np.diag(np.sqrt(d))
        inverse_diagonal = inv(diagonal)
        correlation = inverse_diagonal @ covariance @ inverse_diagonal
        return correlation
    return np.full_like(covariance, np.finfo(float).max)


class EstimationResults:
    """Extension of the raw estimation results."""

    def __init__(self, raw_estimation_results: RawEstimationResults) -> None:
        """Ctor. with data from the object

        :param raw_estimation_results: raw estimation results
        """
        if raw_estimation_results is None:
            raise BiogemeError('Estimation results are missing')
        if not isinstance(raw_estimation_results, RawEstimationResults):
            raise BiogemeError(
                f'Estimation results must be provided as a RawEstimationResults object, '
                f'and not {type(raw_estimation_results)}'
            )
        self.raw_estimation_results = raw_estimation_results

        self._are_derivatives_available = True
        self._is_hessian_available = raw_estimation_results.hessian is not None

        if (
            self.raw_estimation_results.gradient == []
            and self.raw_estimation_results.hessian == [[]]
            and self.raw_estimation_results.bhhh == [[]]
        ):
            self._are_derivatives_available = False
        elif (
            len(self.raw_estimation_results.beta_names)
            != len(self.raw_estimation_results.gradient)
            or (
                self._is_hessian_available
                and (
                    len(self.raw_estimation_results.beta_names)
                    != len(self.raw_estimation_results.hessian)
                )
            )
            or len(self.raw_estimation_results.beta_names)
            != len(self.raw_estimation_results.bhhh)
        ):
            error_msg = 'Incompatible derivatives for {len(self.raw_estimation_results.beta_names)} parameters.'
            raise BiogemeError(error_msg)

    @classmethod
    def from_yaml_file(
        cls,
        filename: str,
    ) -> EstimationResults:
        """Ctor. with data from the file"""
        try:
            restored_results: RawEstimationResults = deserialize_from_yaml(
                filename=filename
            )
        except ConstructorError as e:
            error_msg = f'File {filename}: {e}'
            raise BiogemeError(error_msg) from e
        return cls(raw_estimation_results=restored_results)

    @classmethod
    def from_pickle_file(
        cls,
        filename: str,
    ) -> EstimationResults:
        """Ctor. with data from the file"""
        restored_results: RawEstimationResults = read_pickle_biogeme(filename=filename)
        return cls(raw_estimation_results=restored_results)

    def __getattr__(self, name: str) -> Any:
        # Check raw_estimation_results without triggering __getattr__ recursively
        if (
            'raw_estimation_results' not in self.__dict__
            or self.__dict__['raw_estimation_results'] is None
        ):
            raise BiogemeError(f'Impossible to obtain {name}. No result available.')
        return getattr(self.raw_estimation_results, name)

    @property
    def are_derivatives_available(self) -> bool:
        return self._are_derivatives_available

    @property
    def is_hessian_available(self) -> bool:
        return self._is_hessian_available

    @property
    def number_of_parameters(self) -> int:
        """Number of estimated parameters"""
        if self.raw_estimation_results is None:
            return 0
        return len(self.raw_estimation_results.beta_names)

    @property
    def number_of_free_parameters(self) -> int:
        """This is the number of estimated parameters, minus those that are at
        their bounds
        """
        if self.raw_estimation_results is None:
            return 0
        return sum(
            not self.is_bound_active(parameter_name=beta) for beta in self.beta_names
        )

    @property
    def final_loglikelihood(self) -> float | None:
        """Final log likelihood after estimation"""
        if self.raw_estimation_results is None:
            return None
        return self.raw_estimation_results.final_log_likelihood

    def is_any_bound_active(self, threshold: float = 1.0e-6) -> bool:
        """

        :param threshold: distance below which the bound is considered to be
            active.
        :return: True is any bound constraint is active
        """
        if self.raw_estimation_results is None:
            return False
        number_of_active_constraints = sum(
            self.is_bound_active(parameter_name=beta, threshold=threshold)
            for beta in self.beta_names
        )
        return number_of_active_constraints > 0

    def is_bound_active(self, parameter_name: str, threshold: float = 1.0e-6) -> bool:
        """Check if one of the two bound is 'numerically' active. Being
        numerically active means that the distance between the value of
        the parameter and one of its bounds is below the threshold.

        :param parameter_name: name of the parameter
        :param threshold: distance below which the bound is considered to be
            active.

        :return: True is one of the two bounds is numerically active.

        :raise BiogemeError: if ``threshold`` is negative.
        :raise BiogemeError: if no result is available
        :raise ValueError: if the parameter does not exist

        """
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if threshold < 0:
            raise BiogemeError(f'Threshold ({threshold}) must be non negative')
        index_param = self.beta_names.index(parameter_name)
        if (
            self.lower_bounds[index_param] is not None
            and np.abs(self.beta_values[index_param] - self.lower_bounds[index_param])
            <= threshold
        ):
            return True
        if (
            self.upper_bounds[index_param] is not None
            and np.abs(self.beta_values[index_param] - self.upper_bounds[index_param])
            <= threshold
        ):
            return True
        return False

    def get_beta_values(self, my_betas: list[str] | None = None) -> dict[str, float]:
        """Retrieve the values of the estimated parameters, by names.

        :param my_betas: names of the requested parameters. If None, all
                  available parameters will be reported. Default: None.

        :return: dict containing the values, where the keys are the names.
        """
        if self.raw_estimation_results is None:
            return {}
        if my_betas is not None:
            for requested_beta in my_betas:
                if requested_beta not in self.beta_names:
                    error_msg = f'Unknown parameter {requested_beta}'
                    raise BiogemeError(error_msg)
        beta_values = {
            name: self.beta_values[index]
            for index, name in enumerate(self.beta_names)
            if my_betas is None or name in my_betas
        }
        return beta_values

    def get_betas_for_sensitivity_analysis(
        self,
        my_betas: list[str] | None = None,
        size: int = 100,
        use_bootstrap: bool = True,
        variance_covariance_type: EstimateVarianceCovariance = EstimateVarianceCovariance.ROBUST,
    ) -> list[dict[str, float]]:
        """Generate draws from the distribution of the estimates, for
        sensitivity analysis.

        :param my_betas: names of the parameters for which draws are requested.
        :param size: number of draws. If use_bootstrap is True, the value is
            ignored and a warning is issued. Default: 100.
        :param use_bootstrap: if True, the bootstrap estimates are
                  directly used. The advantage is that it does not rely on the
                  assumption that the estimates follow a normal
                  distribution.
        :param variance_covariance_type: type of variance covariance matrix to use.

        :raise biogeme.exceptions.BiogemeError: if use_bootstrap is True and
            the bootstrap results are not available

        :return: list of dict. Each dict has a many entries as parameters.
                The list has as many entries as draws.

        """
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if my_betas is None:
            my_betas = self.raw_estimation_results.beta_names
        if use_bootstrap and self.raw_estimation_results.bootstrap is None:
            err = (
                'Bootstrap results are not available for simulation. '
                'Use use_bootstrap=False.'
            )
            raise BiogemeError(err)

        index = [self.get_parameter_index(parameter_name=b) for b in my_betas]
        # If we use bootstrap, we simply return the bootstrap draws
        if use_bootstrap:
            results = [
                {my_betas[i]: value for i, value in enumerate([row[j] for j in index])}
                for row in self.raw_estimation_results.bootstrap
            ]
            return results

        the_matrix = self.get_variance_covariance_matrix(
            variance_covariance_type=variance_covariance_type
        )
        simulated_betas = np.random.multivariate_normal(
            self.beta_values, the_matrix, size
        )

        results = [
            {my_betas[i]: value for i, value in enumerate(row)}
            for row in simulated_betas[:, index]
        ]
        return results

    @property
    def akaike_information_criterion(self) -> float:
        """Calculates the AIC

        :return: AIC
        """
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        akaike = (
            2.0 * self.number_of_parameters
            - 2 * self.raw_estimation_results.final_log_likelihood
        )
        return akaike

    @property
    def bayesian_information_criterion(self) -> float:
        """Calculates the BIC

        :return: BIC
        """
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        bayesian = (
            -2.0 * self.raw_estimation_results.final_log_likelihood
            + self.number_of_parameters
            * np.log(self.raw_estimation_results.sample_size)
        )
        return bayesian

    @property
    def algorithm_has_converged(self) -> bool:
        """Reports if the algorithm has indeed converged

        :return: True if the algorithm has converged.
        :rtype: bool
        """
        if self.raw_estimation_results is None:
            return False
        return self.raw_estimation_results.convergence

    @property
    def variance_covariance_missing(self) -> bool:
        """Check if the variance covariance matrix is missing

        :return: True if missing.
        :rtype: bool
        """

        if self.raw_estimation_results is None:
            return True
        return not self.are_derivatives_available

    def calculate_test(self, i: int, j: int, matrix: np.ndarray) -> float:
        """Calculates a t-test comparing two coefficients

        Args:
           i: index of first coefficient \f$\\beta_i\f$.
           j: index of second coefficient \f$\\beta_i\f$.
           matrix: estimate of the variance-covariance matrix \f$m\f$.

        :return: t test
            ..math::  \f[\\frac{\\beta_i-\\beta_j}
                  {\\sqrt{m_{ii}+m_{jj} - 2 m_{ij} }}\f]
        :rtype: float

        """
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        vi = self.raw_estimation_results.beta_values[i]
        vj = self.raw_estimation_results.beta_values[j]
        var_i = matrix[i, i]
        var_j = matrix[j, j]
        covar = matrix[i, j]
        r = var_i + var_j - 2.0 * covar
        if r <= 0:
            test = np.finfo(float).max
        else:
            test = (vi - vj) / np.sqrt(r)
        return test

    @property
    def likelihood_ratio_null(self) -> float:
        """Likelihood ratio test against the null model"""
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if self.raw_estimation_results.null_log_likelihood is None:
            raise BiogemeError('Null log likelihood has not been calculated')
        test = -2.0 * (
            self.raw_estimation_results.null_log_likelihood
            - self.raw_estimation_results.final_log_likelihood
        )
        return test

    @property
    def likelihood_ratio_init(self) -> float:
        """Likelihood ratio test against the initial model"""
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if self.raw_estimation_results.initial_log_likelihood is None:
            raise BiogemeError('Initial log likelihood not available')
        test = -2.0 * (
            self.raw_estimation_results.initial_log_likelihood
            - self.raw_estimation_results.final_log_likelihood
        )
        return test

    @property
    def rho_square_init(self) -> float:
        """McFadden rho square normalized to the initial model"""
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if self.raw_estimation_results.initial_log_likelihood is None:
            raise BiogemeError('Initial log likelihood not available')
        try:
            rho_square = np.nan_to_num(
                1.0
                - self.raw_estimation_results.final_log_likelihood
                / self.raw_estimation_results.initial_log_likelihood
            )
            return rho_square
        except ZeroDivisionError:
            raise BiogemeError('Initial log likelihood is zero')

    @property
    def rho_square_null(self) -> float:
        """McFadden rho square normalized to the null model"""
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if self.raw_estimation_results.null_log_likelihood is None:
            raise BiogemeError('Null log likelihood not available')
        try:
            rho_square = np.nan_to_num(
                1.0
                - self.raw_estimation_results.final_log_likelihood
                / self.raw_estimation_results.null_log_likelihood
            )
            return rho_square
        except ZeroDivisionError:
            raise BiogemeError('Null log likelihood is zero')

    @property
    def rho_bar_square_init(self) -> float:
        """Corrected McFadden rho square normalized to the initial model"""
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if self.raw_estimation_results.initial_log_likelihood is None:
            raise BiogemeError('Initial log likelihood not available')
        try:
            rho_bar_square = np.nan_to_num(
                1.0
                - (
                    self.raw_estimation_results.final_log_likelihood
                    - self.number_of_parameters
                )
                / self.raw_estimation_results.initial_log_likelihood
            )
            return rho_bar_square
        except ZeroDivisionError:
            raise BiogemeError('Initial log likelihood is zero')

    @property
    def rho_bar_square_null(self) -> float:
        """Corrected McFadden rho square normalized to the null model"""
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if self.raw_estimation_results.null_log_likelihood is None:
            raise BiogemeError('Null log likelihood not available')
        try:
            rho_square = np.nan_to_num(
                1.0
                - (
                    self.raw_estimation_results.final_log_likelihood
                    - self.number_of_parameters
                )
                / self.raw_estimation_results.null_log_likelihood
            )
            return rho_square
        except ZeroDivisionError:
            raise BiogemeError('Null log likelihood is zero')

    @property
    def gradient_norm(self) -> float:
        """Norm of the final gradient"""
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available.')
        if not self.are_derivatives_available:
            raise BiogemeError('No gradient available')
        gradient_norm = float(norm(self.raw_estimation_results.gradient))
        return gradient_norm

    def eigen_structure(self) -> tuple[np.ndarray, np.ndarray]:
        """Calculate the eigen structure of the hessian matrix"""
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        if not self.is_hessian_available:
            raise BiogemeError('Second derivative matrix unavailable')
        try:
            eigenvalues, eigenvectors = eigh(
                -np.nan_to_num(np.array(self.raw_estimation_results.hessian))
            )
            return eigenvalues, eigenvectors
        except (LinAlgError, TypeError) as e:
            raise BiogemeError(f'Error in calculating the eigen structure: {e}')

    @property
    def smallest_eigenvalue(self) -> float:
        """Calculates the smallest eigen value of the hessian matrix."""
        eigen_structure = self.eigen_structure()
        eigenvalues, _ = eigen_structure
        return float(np.min(eigenvalues))

    @property
    def smallest_eigenvector(self) -> np.ndarray:
        """Calculates the eigenvector corresponding to the smallest eigen value of the hessian matrix."""
        eigen_structure = self.eigen_structure()
        eigenvalues, eigenvectors = eigen_structure
        min_eigen_index = np.argmin(eigenvalues)
        smallest_eigen_vector = eigenvectors[:, min_eigen_index]
        return smallest_eigen_vector

    @property
    def largest_eigenvalue(self) -> float:
        """Calculates the largest eigen value of the hessian matrix."""
        eigen_structure = self.eigen_structure()
        eigenvalues, _ = eigen_structure
        return float(np.max(eigenvalues))

    @property
    def largest_eigenvector(self) -> np.ndarray:
        """Calculates the eigenvector corresponding to the largest eigen value of the hessian matrix."""
        eigen_structure = self.eigen_structure()
        eigenvalues, eigenvectors = eigen_structure
        max_eigen_index = np.argmax(eigenvalues)
        smallest_eigen_vector = eigenvectors[:, max_eigen_index]
        return smallest_eigen_vector

    @property
    def condition_number(self) -> float:
        """Calculates the condition number of the hessian matrix"""
        try:
            condition_number = self.largest_eigenvalue / self.smallest_eigenvalue
            return condition_number
        except (ZeroDivisionError, TypeError):
            return float(np.finfo(np.float64).max)

    @property
    def rao_cramer_variance_covariance_matrix(self) -> np.ndarray:
        """Calculates the variance-covariance matrix as the (pseudo) inverse of the hessian. We use the pseudo
        inverse in case the matrix is singular

        """
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available.')
        if not self.is_hessian_available:
            raise BiogemeError('Second derivatives matrix not available.')
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        var_covar = -pinv(np.nan_to_num(np.array(self.raw_estimation_results.hessian)))
        return var_covar

    @property
    def bhhh_variance_covariance_matrix(self) -> np.ndarray:
        """Calculates the variance-covariance matrix as the (pseudo) inverse of the BHHH matrix. We use the pseudo
        inverse in case the matrix is singular

        """
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available.')
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        var_covar = pinv(np.nan_to_num(np.array(self.raw_estimation_results.bhhh)))
        return var_covar

    def get_variance_covariance_matrix(
        self, variance_covariance_type: EstimateVarianceCovariance | None
    ) -> np.ndarray:
        """Returns the variance-covariance matrix of a given type"""
        if variance_covariance_type == EstimateVarianceCovariance.BOOTSTRAP:
            return self.bootstrap_variance_covariance_matrix
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        if variance_covariance_type == EstimateVarianceCovariance.BHHH:
            return self.bhhh_variance_covariance_matrix
        if not self.is_hessian_available:
            raise BiogemeError('Second derivatives matrix not available.')
        if variance_covariance_type == EstimateVarianceCovariance.RAO_CRAMER:
            return self.rao_cramer_variance_covariance_matrix
        if variance_covariance_type == EstimateVarianceCovariance.ROBUST:
            return self.robust_variance_covariance_matrix

        raise ValueError(f'Unknown type: {variance_covariance_type}')

    def get_sub_variance_covariance_matrix(
        self,
        parameters: list[str],
        variance_covariance_type: EstimateVarianceCovariance = EstimateVarianceCovariance.ROBUST,
    ) -> np.ndarray:
        """Returns a sub-matrix of the variance-covariance matrix corresponding to a list of parameters

        :param parameters: list of parameters to involve in the sub-matrix
        :param variance_covariance_type: identifies the estimate of the variance-covariance matrix.
        :return: a numpy array containing the submatrix.
        """
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        the_complete_matrix = self.get_variance_covariance_matrix(
            variance_covariance_type=variance_covariance_type
        )
        # Ensure all requested parameters exist in beta_names
        missing_parameters = [
            param for param in parameters if param not in self.beta_names
        ]
        if missing_parameters:
            raise BiogemeError(
                f'The following parameters are unknown: {", ".join(missing_parameters)}'
            )

        # Find indices of the requested parameters
        parameter_indices = [self.beta_names.index(param) for param in parameters]

        # Extract the sub-matrix using NumPy indexing
        sub_matrix = the_complete_matrix[np.ix_(parameter_indices, parameter_indices)]

        return sub_matrix

    @property
    def robust_variance_covariance_matrix(self) -> np.ndarray:
        """Calculates the "sandwich" estimate of the variance-covariance matrix"""
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        robust_var_covar = (
            self.rao_cramer_variance_covariance_matrix
            @ np.array(self.raw_estimation_results.bhhh)
            @ self.rao_cramer_variance_covariance_matrix
        )
        return robust_var_covar

    @property
    def bootstrap_variance_covariance_matrix(self) -> np.ndarray:
        """Calculates the bootstrap variance-covariance matrix"""

        if self.raw_estimation_results.bootstrap is None:
            error_msg = f'No bootstrap samples is available.'
            raise BiogemeError(error_msg)

        bootstrap_var_covar = np.cov(
            np.array(self.raw_estimation_results.bootstrap), rowvar=False
        )
        return bootstrap_var_covar

    def dump_yaml_file(self, filename: str) -> None:
        """Dump the raw estimation results in a Yaml file

        :param filename: name of the file
        """
        if self.raw_estimation_results is None:
            raise BiogemeError('No result available')
        serialize_to_yaml(data=self.raw_estimation_results, filename=filename)

    @property
    def monte_carlo(self) -> bool:
        """Verifies if the model involves Monte-Carlo simulation"""
        if self.raw_estimation_results is None:
            return False
        return self.raw_estimation_results.monte_carlo

    def short_summary(self) -> str:
        """Provides a short summary of the estimation results"""
        if self.raw_estimation_results is None:
            text = 'No estimation result is available.'
            return text

        text = ''
        text += f'Results for model {self.raw_estimation_results.model_name}\n'
        text += f'Nbr of parameters:\t\t{self.number_of_parameters}\n'
        text += f'Sample size:\t\t\t{self.raw_estimation_results.sample_size}\n'
        if (
            self.raw_estimation_results.sample_size
            != self.raw_estimation_results.number_of_observations
        ):
            text += f'Observations:\t\t\t{self.raw_estimation_results.number_of_observations}\n'
        text += f'Excluded data:\t\t\t{self.raw_estimation_results.number_of_excluded_data}\n'
        if self.raw_estimation_results.null_log_likelihood is not None:
            text += f'Null log likelihood:\t\t{self.raw_estimation_results.null_log_likelihood:.7g}\n'
        text += f'Final log likelihood:\t\t{self.raw_estimation_results.final_log_likelihood:.7g}\n'
        if self.raw_estimation_results.null_log_likelihood is not None:
            text += (
                f'Likelihood ratio test (null):\t\t'
                f'{self.likelihood_ratio_null:.7g}\n'
            )
            text += f'Rho square (null):\t\t\t{self.rho_square_null:.3g}\n'
            text += f'Rho bar square (null):\t\t\t' f'{self.rho_bar_square_null:.3g}\n'
        text += (
            f'Akaike Information Criterion:\t{self.akaike_information_criterion:.7g}\n'
        )
        text += f'Bayesian Information Criterion:\t{self.bayesian_information_criterion:.7g}\n'
        return text

    def __str__(self) -> str:
        return self.short_summary()

    def get_general_statistics(self) -> dict[str, str]:
        """Format the results in a dict

        :return: dict with the results. The keys describe each
                 content.

        """
        d = {'Number of estimated parameters': f'{self.number_of_parameters}'}
        if self.number_of_free_parameters != self.number_of_parameters:
            d['Number of free parameters'] = f'{self.number_of_free_parameters}'
        d['Sample size'] = f'{self.sample_size}'
        if self.sample_size != self.number_of_observations:
            d['Observations'] = f'{self.number_of_observations}'
        d['Excluded observations'] = f'{self.number_of_excluded_data}'
        if self.null_log_likelihood is not None:
            d['Null log likelihood'] = f'{self.null_log_likelihood:.7g}'
        if self.initial_log_likelihood is not None:
            d['Init log likelihood'] = f'{self.initial_log_likelihood:.7g}'
        d['Final log likelihood'] = f'{self.final_log_likelihood:.7g}'
        if self.null_log_likelihood is not None:
            d['Likelihood ratio test for the null model'] = (
                f'{self.likelihood_ratio_null:.7g}'
            )
            d['Rho-square for the null model'] = f'{self.rho_square_null:.3g}'
            d['Rho-square-bar for the null model'] = f'{self.rho_bar_square_null:.3g}'
        try:
            likelihood_ratio_init = f'{self.likelihood_ratio_init:.7g}'
        except (BiogemeError, ZeroDivisionError):
            likelihood_ratio_init = ''
        d['Likelihood ratio test for the init. model'] = likelihood_ratio_init
        try:
            rho_square_init = f'{self.rho_square_init:.3g}'
        except (BiogemeError, ZeroDivisionError):
            rho_square_init = ''
        d['Rho-square for the init. model'] = rho_square_init
        try:
            rho_bar_square_init = f'{self.rho_bar_square_init:.3g}'
        except (BiogemeError, ZeroDivisionError):
            rho_bar_square_init = ''
        d['Rho-square-bar for the init. model'] = rho_bar_square_init
        try:
            akaike = f'{self.akaike_information_criterion:.7g}'
        except BiogemeError:
            akaike = ''
        d['Akaike Information Criterion'] = akaike
        try:
            bayesian = f'{self.bayesian_information_criterion:.7g}'
        except BiogemeError:
            bayesian = ''
        d['Bayesian Information Criterion'] = bayesian

        try:
            gradient_norm = f'{self.gradient_norm:.4E}'
        except BiogemeError:
            gradient_norm = ''
        d['Final gradient norm'] = gradient_norm
        if self.monte_carlo:
            d['Number of draws'] = f'{self.number_of_draws}'
            d['Draws generation time'] = f'{self.draws_processing_time}'
            d['Types of draws'] = ', '.join(
                [f'{i}: {k}' for i, k in self.types_of_draws.items()]
            )
        if self.bootstrap is not None:
            d['Bootstrapping time'] = f'{self.bootstrap_time}'
        return d

    def get_parameter_index(self, parameter_name: str) -> int:
        """Retrieve the index of a parameter

        :param parameter_name: name of the parameter
        :return: index of the parameter
        """
        index_param = self.beta_names.index(parameter_name)
        return index_param

    def get_parameter_value_from_index(self, parameter_index: int) -> float:
        """Retrieve the estimated value of a parameter

        :param parameter_index: index of the parameter
        """
        if parameter_index < 0 or parameter_index >= len(self.beta_values):
            error_msg = (
                f'Invalid parameter index {parameter_index}. Valid range: 0- '
                f'{len(self.beta_values)-1}'
            )
            raise ValueError(error_msg)

        return self.beta_values[parameter_index]

    def get_parameter_value(self, parameter_name: str) -> float:
        """Retrieve the estimated value of a parameter

        :param parameter_name: name of the parameter
        """
        index = self.get_parameter_index(parameter_name=parameter_name)
        return self.get_parameter_value_from_index(parameter_index=index)

    def get_parameter_std_err_from_index(
        self,
        parameter_index: int,
        estimate_var_covar: EstimateVarianceCovariance | None = None,
    ) -> float:
        if estimate_var_covar is None:
            estimate_var_covar = self.get_default_variance_covariance_matrix()
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        """Calculates the standard error of the parameter estimate"""
        if parameter_index < 0 or parameter_index >= len(self.beta_values):
            error_msg = (
                f'Invalid parameter index {parameter_index}. Valid range: 0- '
                f'{len(self.beta_values)-1}'
            )
            raise ValueError(error_msg)
        var_covar = self.get_variance_covariance_matrix(
            variance_covariance_type=estimate_var_covar
        )
        variance = var_covar[parameter_index, parameter_index]
        if variance < 0:
            return float(np.finfo(float).max)
        return np.sqrt(variance)

    def get_parameter_std_err(
        self,
        parameter_name: str,
        estimate_var_covar: EstimateVarianceCovariance | None = None,
    ) -> float:
        if estimate_var_covar is None:
            estimate_var_covar = self.get_default_variance_covariance_matrix()
        """Calculates the standard error of the parameter estimate"""
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        index = self.get_parameter_index(parameter_name=parameter_name)
        return self.get_parameter_std_err_from_index(
            parameter_index=index, estimate_var_covar=estimate_var_covar
        )

    def get_parameter_t_test_from_index(
        self,
        parameter_index: int,
        estimate_var_covar: EstimateVarianceCovariance | None = None,
        target: float = 0.0,
    ) -> float:
        """Calculates the t-test of the parameter estimate

        :param parameter_index: index of the parameter
        :param estimate_var_covar: estimator for the variance-covariance matrix
        :param target: value for the null hypothesis
        :return: value of the t-test
        """
        if estimate_var_covar is None:
            estimate_var_covar = self.get_default_variance_covariance_matrix()
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        if parameter_index < 0 or parameter_index >= len(self.beta_values):
            error_msg = (
                f'Invalid parameter index {parameter_index}. Valid range: 0- '
                f'{len(self.beta_values)-1}'
            )
            raise ValueError(error_msg)
        value = self.get_parameter_value_from_index(parameter_index=parameter_index)
        std_err = self.get_parameter_std_err_from_index(
            parameter_index=parameter_index, estimate_var_covar=estimate_var_covar
        )
        if std_err == 0:
            return float(np.finfo(float).max)

        return float(np.nan_to_num((value - target) / std_err))

    def get_parameter_t_test(
        self,
        parameter_name: str,
        estimate_var_covar: EstimateVarianceCovariance | None = None,
        target: float = 0.0,
    ) -> float:
        """Calculates the t-test of the parameter estimate

        :param parameter_name: name of the parameter
        :param estimate_var_covar: estimator for the variance-covariance matrix
        :param target: value for the null hypothesis
        :return: value of the t-test
        """
        if estimate_var_covar is None:
            estimate_var_covar = self.get_default_variance_covariance_matrix()
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        index = self.get_parameter_index(parameter_name=parameter_name)
        return self.get_parameter_t_test_from_index(
            parameter_index=index, estimate_var_covar=estimate_var_covar, target=target
        )

    def get_parameter_p_value_from_index(
        self,
        parameter_index: int,
        estimate_var_covar: EstimateVarianceCovariance | None = None,
        target: float = 0.0,
    ) -> float:
        """Calculates the p-value of the parameter estimate

        :param parameter_index: index of the parameter
        :param estimate_var_covar: estimator for the variance-covariance matrix
        :param target: value for the null hypothesis
        :return: p-value
        """
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        if estimate_var_covar is None:
            estimate_var_covar = self.get_default_variance_covariance_matrix()
        if parameter_index < 0 or parameter_index >= len(self.beta_values):
            error_msg = (
                f'Invalid parameter index {parameter_index}. Valid range: 0- '
                f'{len(self.beta_values)-1}'
            )
            raise ValueError(error_msg)
        t_test = self.get_parameter_t_test_from_index(
            parameter_index=parameter_index,
            estimate_var_covar=estimate_var_covar,
            target=target,
        )
        return calc_p_value(t_test)

    def get_parameter_p_value(
        self,
        parameter_name: str,
        estimate_var_covar: EstimateVarianceCovariance | None = None,
        target: float = 0.0,
    ) -> float:
        """Calculates the p-value of the parameter estimate

        :param parameter_name: name of the parameter
        :param estimate_var_covar: estimator for the variance-covariance matrix
        :param target: value for the null hypothesis
        :return: p-value
        """
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        if estimate_var_covar is None:
            estimate_var_covar = self.get_default_variance_covariance_matrix()
        index = self.get_parameter_index(parameter_name=parameter_name)
        return self.get_parameter_p_value_from_index(
            parameter_index=index, estimate_var_covar=estimate_var_covar, target=target
        )

    def likelihood_ratio_test(
        self, other_model: EstimationResults, significance_level: float = 0.05
    ) -> likelihood_ratio.LRTuple:
        """This function performs a likelihood ratio test between a restricted
        and an unrestricted model. The "self" model can be either the
        restricted or the unrestricted.

        :param other_model: other model to perform the test.
        :param significance_level: level of significance of the
            test. Default: 0.05
        :return: a tuple containing:

                  - a message with the outcome of the test
                  - the statistic, that is minus two times the difference
                    between the loglikelihood  of the two models
                  - the threshold of the chi square distribution.

        """
        lr = self.final_log_likelihood
        lu = other_model.final_log_likelihood
        kr = self.number_of_parameters
        ku = other_model.number_of_parameters
        return likelihood_ratio.likelihood_ratio_test(
            (lu, ku), (lr, kr), significance_level
        )

    def get_estimated_parameters(self) -> pd.DataFrame:
        """For backward compatibility"""
        from . import get_pandas_estimated_parameters

        msg = (
            'get_estimated_parameters is deprecated. '
            'Use get_pandas_estimated_parameters(estimation_results=my_results) instead'
        )
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return get_pandas_estimated_parameters(estimation_results=self)

    def get_confidence_ellipse(
        self,
        first_parameter: str,
        second_parameter: str,
        variance_covariance_type: EstimateVarianceCovariance | None = None,
        confidence_level=0.95,
    ) -> Ellipse:
        """Provides a Tikz picture of the confidence ellipsis for two parameters
        :param first_parameter: name of the first parameter
        :param second_parameter: name of the second parameter
        :param variance_covariance_type: type of variance-covariance estimate to be used.
        :param confidence_level: level of confidence for the confidence ellipse
        :return: LaTeX code to draw the ellipsis
        """
        if not self.are_derivatives_available:
            raise BiogemeError('No derivatives available.')
        if variance_covariance_type is None:
            variance_covariance_type = self.get_default_variance_covariance_matrix()
        sigma = self.get_sub_variance_covariance_matrix(
            variance_covariance_type=variance_covariance_type,
            parameters=[first_parameter, second_parameter],
        )
        beta_1 = self.get_parameter_value_from_index(
            parameter_index=self.get_parameter_index(parameter_name=first_parameter)
        )
        beta_2 = self.get_parameter_value_from_index(
            parameter_index=self.get_parameter_index(parameter_name=second_parameter)
        )
        # Mean (center of the ellipse)
        center_x, center_y = beta_1, beta_2

        # Eigenvalue decomposition
        eigenvalues, eigenvectors = np.linalg.eigh(sigma)
        # Extract eigenvector corresponding to the largest eigenvalue
        largest_eigenvector = eigenvectors[:, np.argmax(eigenvalues)]
        # Calculate cos_phi and sin_phi
        cos_phi, sin_phi = largest_eigenvector

        # Compute axis lengths
        chi2_value = chi2.ppf(confidence_level, 2)
        axis_lengths = 2 * np.sqrt(chi2_value * eigenvalues)
        axis_one, axis_two = axis_lengths[1], axis_lengths[0]  # Major and minor axes
        return Ellipse(
            center_x=center_x,
            center_y=center_y,
            sin_phi=sin_phi,
            cos_phi=cos_phi,
            axis_one=axis_one,
            axis_two=axis_two,
        )

    @deprecated_parameters({'robust_std_err': None})
    def write_f12(self, overwrite=False) -> str:
        """
        Writes the estimation results to a file in the F12 format (ALOGIT).

        :param overwrite: If True, the existing file will be overwritten. Default is False.
        :return: name of the file
        """
        from .f12_output import generate_f12_file

        filename = get_new_file_name(self.model_name, 'F12')
        generate_f12_file(self, filename, overwrite=overwrite)
        return filename

    def write_latex(
        self,
        variance_covariance_type: EstimateVarianceCovariance | None = None,
        include_begin_document=False,
    ) -> str:
        """Write the results in a LaTeX file."""
        from .latex_output import generate_latex_file

        if variance_covariance_type is None:
            variance_covariance_type = self.get_default_variance_covariance_matrix()

        latex_file_name = get_new_file_name(self.model_name, 'tex')
        generate_latex_file(
            estimation_results=self,
            filename=latex_file_name,
            include_begin_document=include_begin_document,
            variance_covariance_type=variance_covariance_type,
        )
        return latex_file_name

    def get_default_variance_covariance_matrix(self) -> EstimateVarianceCovariance:
        """Selects the default variance covariance matrix"""

        if self.bootstrap_time is not None:
            return EstimateVarianceCovariance.BOOTSTRAP

        if self.is_hessian_available:
            return EstimateVarianceCovariance.ROBUST

        return EstimateVarianceCovariance.BHHH



"""

biogeme.models
==============

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Wed Nov 22 15:24:34 2023
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from biogeme.version import getText
import biogeme.database as db
from biogeme import models
from biogeme.nests import (
    OneNestForNestedLogit,
    OneNestForCrossNestedLogit,
    NestsForNestedLogit,
    NestsForCrossNestedLogit,
)
from biogeme.expressions import Variable, Beta

# %%
# Version of Biogeme.
print(getText())


# %%
# Definition of a database
# ++++++++++++++++++++++++
df = pd.DataFrame(
    {
        'Person': [1, 1, 1, 2, 2],
        'Exclude': [0, 0, 1, 0, 1],
        'Variable1': [1, 2, 3, 4, 5],
        'Variable2': [10, 20, 30, 40, 50],
        'Choice': [1, 2, 3, 1, 2],
        'Av1': [0, 1, 1, 1, 1],
        'Av2': [1, 1, 0, 1, 1],
        'Av3': [0, 1, 1, 1, 0],
    }
)
df

# %%
my_data = db.Database('test', df)

# %%
# Piecewise linear specification
# ++++++++++++++++++++++++++++++

# %%
# A piecewise linear specification (sometimes called 'spline') is a
# continuous but not differentiable function of the variable. It is
# defined based on thresholds. Between two thresholds, the function is
# linear. And the slope is changing after each threshold.
#
# Consider a variable :math:`t` and an interval :math:`[a, a+b]`. We
# define a new variable
#
# .. math:: x_{[a,b]}(t) = \max(0,\min(t-a,b)) =
#     \left\{
#        \begin{array}{ll}
#            0 & \text{if } t < a, \\
#            t-a & \text{if } a \leq t < a+b, \\
#            b  & \text{otherwise}.
#        \end{array}
#     \right.
#
# For each interval :math:`]-\infty,a]`, we have
#
# .. math:: x_{]-\infty,a]}(t) = \min(t,a) = \left\{
#     \begin{array}{ll}
#         t & \text{if } t < a, \\
#         a  & \text{otherwise}.
#      \end{array}
#      \right..
#
# For each interval :math:`[a,+\infty[`, we have
#
# .. math:: x_{]-\infty,a]}(t) = \max(0,t-a) = \left\{
#     \begin{array}{ll} 0& \text{if } t < a, \\ t-a &
#     \text{otherwise}.  \end{array} \right..
#
# If we consider a series of threshold
#
# .. math:: \alpha_1 < \alpha_2 < \ldots <\alpha_K,
#
# the piecewise linear transform of variable :math:`t` is
#
# .. math:: \sum_{k=1}^{K-1} \beta_k x_{[\alpha_k,\alpha_{k+1}]},
#
# where :math:`\beta_k` is the slope of the linear function in interval
# :math:`[\alpha_k,\alpha_{k+1}]`.
#

# %%
# The next statement generates the variables, given the thresholds. A
# `None` is equivalent to :math:`\infty`, and can only appear first (and it
# means :math:`-\infty`) or last (and it means :math:`+\infty`).
x = Variable('x')
thresholds = [None, 90, 180, 270, None]
variables = models.piecewiseVariables(x, thresholds)
print(variables)

# %%
# The next statement automatically generates the formula, including
# the Beta parameters, that are initialized to zero.
formula = models.piecewiseFormula('x', thresholds)
print(formula)

# %%
# It is also possible to initialize the Beta parameters with other
# values. Note also that the first argument can be either the name of
# the variable (as in the previous call) or the variable itself.
betas = [-0.016806308, -0.010491137, -0.002012234, -0.020051303]
formula = models.piecewiseFormula(x, thresholds, betas)
print(formula)

# %%
# We provide a plot of a piecewise linear specification.

# %%
X = np.arange(0, 300, 0.1)
Y = [
    models.piecewiseFunction(
        x, thresholds, [-0.016806308, -0.010491137, -0.002012234, -0.020051303]
    )
    for x in X
]
plt.plot(X, Y)

# %%
# Logit
# +++++

# %%
V = {1: Variable('Variable1'), 2: 0.1, 3: -0.1}
av = {1: Variable('Av1'), 2: Variable('Av2'), 3: Variable('Av3')}

# %%
# Calculation of the (log of the) logit for the three alternatives,
# based on their availability.

# %%
# Alternative 1
p1 = models.logit(V, av, 1)
p1.getValue_c(my_data, prepareIds=True)

# %%
p1 = models.loglogit(V, av, 1)
p1.getValue_c(my_data, prepareIds=True)

# %%
# Alternative 2
p2 = models.logit(V, av, 2)
p2.getValue_c(my_data, prepareIds=True)

# %%
p2 = models.loglogit(V, av, 2)
p2.getValue_c(my_data, prepareIds=True)

# %%
# Alternative 3
p3 = models.logit(V, av, 3)
p3.getValue_c(my_data, prepareIds=True)

# %%
p3 = models.loglogit(V, av, 3)
p3.getValue_c(my_data, prepareIds=True)

# %%
# Calculation of the log of the logit for the three alternatives,
# **assuming that they are all available**.

# %%
# Alternative 1
pa1 = models.logit(V, av=None, i=1)
pa1.getValue_c(my_data, prepareIds=True)

# %%
pa1 = models.loglogit(V, av=None, i=1)
pa1.getValue_c(my_data, prepareIds=True)

# %%
# Alternative 2
pa2 = models.logit(V, av=None, i=2)
pa2.getValue_c(my_data, prepareIds=True)

# %%
pa2 = models.loglogit(V, av=None, i=2)
pa2.getValue_c(my_data, prepareIds=True)

# %%
# Alternative 3
pa3 = models.logit(V, av=None, i=3)
pa3.getValue_c(my_data, prepareIds=True)

# %%
pa3 = models.loglogit(V, av=None, i=3)
pa3.getValue_c(my_data, prepareIds=True)

# %%
# Boxcox transform
# ++++++++++++++++

# %%
# The Box-Cox transform of a variable :math:`x` is defined as
#
# .. math:: B(x,\ell) =\frac{x^{\ell}-1}{\ell},
#
# where :math:`\ell` is a parameter that can be estimated from data.  It has
# the property that
#
# .. math:: \lim_{\ell \to 0} B(x,\ell)=\log(x).
x = Variable('Variable1')
models.boxcox(x, 4)

# %%
x = Variable('Variable1')
models.boxcox(x, 0)

# %%
ell = Variable('Variable2')
e = models.boxcox(x, ell)
print(e)

# %%
e.getValue_c(my_data, prepareIds=True)

# %%
# We numerically illustrate that, when :math:`\lambda` goes to 0, the BoxCox
# transform of :math:`x` converges to the log of :math:`x`.

# %%
for ell in range(1, 16):
    x = 3
    bc = models.boxcox(x, 10**-ell).getValue()
    print(f'ell=l0^(-{ell}): {bc:.6g} - {np.log(x):.6g} = {bc - np.log(x):.6g}')


# %%
# MEV models
# ++++++++++

# %%
# MEV models are defined as
#
# .. math:: \frac{e^{V_i + \ln G_i(e^{V_1},\ldots,e^{V_J})}}{\sum_j
#     e^{V_j + \ln G_j(e^{V_1},\ldots,e^{V_J})}},
#
# where :math:`G` is a generating function, and
#
# .. math:: G_i=\frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J}).

# %%
# **Nested logit model**:
# the :math:`G` function for the nested logit model is defined such that
#
# .. math:: G_i=\frac{\partial G}{\partial
#     y_i}(e^{V_1},\ldots,e^{V_J}) = \mu e^{(\mu_m-1)V_i}
#     \left(\sum_{i=1}^{J_m} e^{\mu_m
#     V_i}\right)^{\frac{\mu}{\mu_m}-1},
#
# where the choice set is partitioned into :math:`J_m` nests, each
# associated with a parameter :math:`\mu_m`, and :math:`\mu` is the scale
# parameter. The condition is :math:`0 \leq \mu \leq \mu_m` must be
# verified. In general, :math:`\mu` is normalized to 1.0.

# %%
# This is an example with 5 alternatives. Nest A contains alternatives
# 1, 2 and 4, and is associated with a scale parameter
# :math:`\mu_A=1.2`. Nest B contains alternatives 3 and 5, and is associated
# with a scale parameter :math:`\mu_B=2.3`.

# %%
V = {1: Variable('Variable1'), 2: 0.1, 3: -0.1, 4: -0.2, 5: 0.2}
av = {1: 1, 2: 0, 3: 1, 4: 1, 5: 1}

# %%
# Definition of the nests.
nest_a = OneNestForNestedLogit(
    nest_param=1.2, list_of_alternatives=[1, 2, 4], name='nest_a'
)
nest_b = OneNestForNestedLogit(
    nest_param=2.3, list_of_alternatives=[3, 5], name='name_b'
)

nests = NestsForNestedLogit(choice_set=list(V), tuple_of_nests=(nest_a, nest_b))

# %%
p1 = models.nested(V, availability=av, nests=nests, choice=1)
p1.getValue_c(my_data, prepareIds=True)

# %%
# If all the alternatives are available, define the availability dictionary as None.
p1 = models.nested(V, availability=None, nests=nests, choice=1)
p1.getValue_c(my_data, prepareIds=True)

# %%
# The syntax is similar to obtain the log of the probability.
p2 = models.lognested(V, availability=av, nests=nests, choice=1)
p2.getValue_c(my_data, prepareIds=True)

# %%
p2 = models.lognested(V, availability=None, nests=nests, choice=1)
p2.getValue_c(my_data, prepareIds=True)

# %%
# If the value of the parameter :math:`\mu` is not 1, there is another
# function to call. Note that, for the sake of computational
# efficiency, it is not verified by the code if the condition :math:`0 \leq
# \mu \leq \mu_m` is valid.
p1 = models.nestedMevMu(V, availability=av, nests=nests, choice=1, mu=1.1)
p1.getValue_c(my_data, prepareIds=True)

# %%
p1 = models.nestedMevMu(V, availability=None, nests=nests, choice=1, mu=1.1)
p1.getValue_c(my_data, prepareIds=True)

# %%
p1 = models.lognestedMevMu(V, availability=av, nests=nests, choice=1, mu=1.1)
p1.getValue_c(my_data, prepareIds=True)

# %%
p1 = models.lognestedMevMu(V, availability=None, nests=nests, choice=1, mu=1.1)
p1.getValue_c(my_data, prepareIds=True)

# %%
# The validity of the nested structure can be verified.
nest_c = OneNestForNestedLogit(nest_param=2.3, list_of_alternatives=[3], name='name_c')
nests = NestsForNestedLogit(choice_set=list(V), tuple_of_nests=(nest_a, nest_b))

is_valid, msg = nests.check_partition()

# %%
is_valid

# %%
print(msg)

# %%
# If an alternative belongs to two nests

# %%
nest_a = OneNestForNestedLogit(nest_param=1.2, list_of_alternatives=[1, 2, 3, 4])
nest_b = OneNestForNestedLogit(nest_param=2.3, list_of_alternatives=[3, 5])
nests = NestsForNestedLogit(choice_set=list(V), tuple_of_nests=(nest_a, nest_b))
is_valid, msg = nests.check_partition()

# %%
is_valid

# %%
print(msg)

# %%
# **Cross-nested logit model**: the :math:`G` function for the cross
# nested logit model is defined such that
#
# .. math:: G_i=\frac{\partial G}{\partial
#     y_i}(e^{V_1},\ldots,e^{V_J}) = \mu \sum_{m=1}^{M}
#     \alpha_{im}^{\frac{\mu_m}{\mu}} e^{(\mu_m-1) V_i}\left(
#     \sum_{j=1}^{J} \alpha_{jm}^{\frac{\mu_m}{\mu}} e^{\mu_m V_j}
#     \right)^{\frac{\mu}{\mu_m}-1},
#
# where each nest :math:`m`  is associated with a parameter :math:`\mu_m` and, for
# each alternative :math:`i`, a parameter :math:`\alpha_{im} \geq 0` that captures
# the degree of membership of alternative :math:`i` to nest :math:`m`. :math:`\mu` is
# the scale parameter. For each alternative :math:`i`, there must be at
# least one nest :math:`m` such that :math:`\alpha_{im}>0`. The condition :math:`0
# \leq \mu \leq \mu_m` must be also verified. In general, :math:`\mu` is
# normalized to 1.0.

# %%
# This is an example with 5 alternatives and two nests.
#
# - Alt. 1 belongs to nest A.
# - Alt. 2 belongs to nest A.
# - Alt. 3 belongs to both nest A (50%) and nest B (50%).
# - Alt. 4 belongs to nest B.
# - Alt. 5 belongs to nest B.

# %%
V = {1: Variable('Variable1'), 2: 0.1, 3: -0.1, 4: -0.2, 5: 0.2}
av = {1: 1, 2: 0, 3: 1, 4: 1, 5: 1}
alpha_a = {1: 1, 2: 1, 3: 0.5, 4: 0, 5: 0}
alpha_b = {1: 0, 2: 0, 3: 0.5, 4: 1, 5: 1}
nest_a = OneNestForCrossNestedLogit(
    nest_param=1.2, dict_of_alpha=alpha_a, name='Nest a'
)
nest_b = OneNestForCrossNestedLogit(
    nest_param=2.3, dict_of_alpha=alpha_b, name='Nest b'
)
nests = NestsForCrossNestedLogit(choice_set=list(V), tuple_of_nests=(nest_a, nest_b))

# %%
p1 = models.cnl(V, availability=av, nests=nests, choice=1)
p1.getValue_c(my_data, prepareIds=True)

# %%
# If all the alternatives are available, define the availability dictionary as None.
p1 = models.cnl(V, availability=None, nests=nests, choice=1)
p1.getValue_c(my_data, prepareIds=True)

# %%
# If the value of the parameter :math:`\mu` is not 1, there is another
# function to call. Note that, for the sake of computational
# efficiency, it is not verified by the code if the condition :math:`0 \leq
# \mu \leq \mu_m` is verified.
p1 = models.cnlmu(V, availability=av, nests=nests, choice=1, mu=1.1)
p1.getValue_c(my_data, prepareIds=True)

# %%
p1 = models.cnlmu(V, availability=None, nests=nests, choice=1, mu=1.1)
p1.getValue_c(my_data, prepareIds=True)

# %%
# If the sample is endogenous, a correction must be included in the
# model, as proposed by `Bierlaire, Bolduc and McFadden (2008)
# <http://dx.doi.org/10.1016/j.trb.2007.09.003>`_.
# In this case, the generating function must first be defined, and the
# MEV model with correction is then called.
logGi = models.getMevForCrossNested(V, availability=av, nests=nests)
logGi

# %%
correction = {1: -0.1, 2: 0.1, 3: 0.2, 4: -0.2, 5: 0}
p1 = models.mev_endogenousSampling(V, logGi, av, correction, choice=1)
p1.getValue_c(my_data, prepareIds=True)

# %%
correction = {1: -0.1, 2: 0.1, 3: 0.2, 4: -0.2, 5: 0}
p1 = models.logmev_endogenousSampling(V, logGi, av, correction, choice=1)
p1.getValue_c(my_data, prepareIds=True)

# %%
correction = {1: -0.1, 2: 0.1, 3: 0.2, 4: -0.2, 5: 0}
p1 = models.mev_endogenousSampling(V, logGi, av=None, correction=correction, choice=1)
p1.getValue_c(my_data, prepareIds=True)

# %%
correction = {1: -0.1, 2: 0.1, 3: 0.2, 4: -0.2, 5: 0}
p1 = models.logmev_endogenousSampling(
    V, logGi, av=None, correction=correction, choice=1
)
p1.getValue_c(my_data, prepareIds=True)

# %%
# The MEV generating function for the following models are available.

# %%
# Nested logit model
V = {1: Variable('Variable1'), 2: 0.1, 3: -0.1, 4: -0.2, 5: 0.2}
av = {1: 1, 2: 0, 3: 1, 4: 1, 5: 1}
nest_a = OneNestForNestedLogit(
    nest_param=Beta('muA', 1.2, 1.0, None, 0), list_of_alternatives=[1, 2, 4]
)
nest_b = OneNestForNestedLogit(
    nest_param=Beta('muB', 2.3, 1.0, None, 0), list_of_alternatives=[3, 5]
)
nests = NestsForNestedLogit(choice_set=list(V), tuple_of_nests=(nest_a, nest_b))

# %%
logGi = models.getMevForNested(V, availability=None, nests=nests)
logGi

# %%
# And with the :math:`\mu` parameter.

# %%
logGi = models.getMevForNestedMu(V, availability=None, nests=nests, mu=1.1)
logGi

# %%
# Cross nested logit model

# %%
V = {1: Variable('Variable1'), 2: 0.1, 3: -0.1, 4: -0.2, 5: 0.2}
av = {1: 1, 2: 0, 3: 1, 4: 1, 5: 1}
alpha_a = {1: 1, 2: 1, 3: 0.5, 4: 0, 5: 0}
alpha_b = {1: 0, 2: 0, 3: 0.5, 4: 1, 5: 1}
nest_a = OneNestForCrossNestedLogit(
    nest_param=Beta('muA', 1.2, 1.0, None, 0), dict_of_alpha=alpha_a
)
nest_b = OneNestForCrossNestedLogit(
    nest_param=Beta('muB', 2.3, 1.0, None, 0), dict_of_alpha=alpha_b
)
nests = NestsForCrossNestedLogit(choice_set=list(V), tuple_of_nests=(nest_a, nest_b))
# %%
logGi = models.getMevForCrossNested(V, availability=None, nests=nests)
logGi

# %%
# Cross nested logit model with :math:`\mu` parameter.

# %%
logGi = models.getMevForCrossNestedMu(V, availability=None, nests=nests, mu=1.1)
logGi



from .compilation import compile_estimation_results
from .estimation_results import (
    EstimateVarianceCovariance,
    EstimationResults,
    calc_p_value,
)
from .f12_output import generate_f12_file, get_f12
from .html_output import (
    format_real_number,
    generate_html_file,
    get_html_condition_number,
    get_html_correlation_results,
    get_html_estimated_parameters,
    get_html_footer,
    get_html_general_statistics,
    get_html_header,
    get_html_one_parameter,
    get_html_preamble,
)
from .latex_output import (
    compare_parameters,
    get_latex_estimated_parameters,
    get_latex_general_statistics,
)
from .pandas_output import (
    get_pandas_correlation_results,
    get_pandas_estimated_parameters,
    get_pandas_one_pair_of_parameters,
    get_pandas_one_parameter,
)
from .pareto import pareto_optimal
from .raw_estimation_results import (
    RawEstimationResults,
    deserialize_from_yaml,
    serialize_to_yaml,
)
from .recycle_pickle import (
    pickle_to_yaml,
    read_pickle_biogeme,
    read_pickle_biogeme_3_2_10,
    read_pickle_biogeme_3_2_11,
    read_pickle_biogeme_3_2_12,
    read_pickle_biogeme_3_2_13,
    read_pickle_biogeme_3_2_14,
    read_pickle_biogeme_3_2_7,
    read_pickle_biogeme_3_2_8,
)
from .variance_covariance import EstimateVarianceCovariance



"""Arithmetic expressions accepted by Biogeme: numeric expressions

Michel Bierlaire
Tue Mar 25 18:41:06 2025
"""

from __future__ import annotations

import logging

import jax.numpy as jnp

from .base_expressions import Expression
from .jax_utils import JaxFunctionType
from ..floating_point import JAX_FLOAT

logger = logging.getLogger(__name__)


class Numeric(Expression):
    """
    Numerical expression for a simple number
    """

    def __init__(self, value: float | int | bool):
        """Constructor

        :param value: numerical value
        :type value: float
        """
        super().__init__()
        self.value = float(value)  #: numeric value

    def deep_flat_copy(self) -> Numeric:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        return type(self)(value=self.value)

    def __str__(self) -> str:
        return '`' + str(self.value) + '`'

    def __repr__(self):
        return f'<Numeric value={self.value}>'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        return self.value

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            return jnp.array(self.value, dtype=JAX_FLOAT)

        return the_jax_function



"""Shortcut to perform simulation of a choice model

Michel Bierlaire
Sat Jan 18 11:24:47 2025
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pandas import DataFrame

from biogeme.expressions import Expression

if TYPE_CHECKING:
    from biogeme.database import Database
    from biogeme.results_processing import EstimationResults

logger = logging.getLogger(__name__)


def simulate(
    database: Database,
    dict_of_expressions: dict[str, Expression],
    estimation_results: EstimationResults,
    csv_filename: str | None = None,
) -> DataFrame:
    """Simulate a discrete choice model on a database.

    :param database: database
    :param dict_of_expressions: expressions that will be simulated
    :param estimation_results: estimation results.
    :param csv_filename: Name of the output CSV file. If None, no file is generated.
    :return: a pandas data frame with the simulated value. Each row corresponds to a row in the database, and each
              column to a formula.
    """
    from biogeme.biogeme import BIOGEME

    biosim = BIOGEME(database, dict_of_expressions)
    biosim.model_name = database.name + '_' + '_'.join(dict_of_expressions.keys())
    simulation_results = biosim.simulate(
        the_beta_values=estimation_results.get_beta_values()
    )
    if csv_filename is not None:
        simulation_results.to_csv(csv_filename, index=False)
        logger.info(f'File {csv_filename} has been generated.')
    return simulation_results



"""

biogeme.draws
=============

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Tue Nov 21 18:36:59 2023
"""
# %%
import numpy as np
import pandas as pd
from biogeme.version import getText
import biogeme.draws as dr

# %%
# Version of Biogeme.
print(getText())

# %%
# We set the seed so that the outcome of random operations is always the same.
np.random.seed(90267)

# %%
# Uniform draws
# -------------

# %%
# Uniform [0,1]. The output is transformed into a data frame just for the display.
draws = dr.getUniform(sample_size=3, number_of_draws=10, symmetric=False)
pd.DataFrame(draws)

# %%
draws = dr.getUniform(sample_size=3, number_of_draws=10, symmetric=True)
pd.DataFrame(draws)

# %%
# LatinHypercube: the Modified Latin Hypercube Sampling (MLHS, Hess et al, 2006)
# provides U[0,1] draws from a perturbed grid, designed for
# Monte-Carlo integration.

# %%
latin_hypercube = dr.getLatinHypercubeDraws(sample_size=3, number_of_draws=10)
pd.DataFrame(latin_hypercube)

# %%
# The same method can be used to generate draws from U[-1,1]

# %%
latin_hypercube = dr.getLatinHypercubeDraws(
    sample_size=5, number_of_draws=10, symmetric=True
)
pd.DataFrame(latin_hypercube)

# %%
# The user can provide her own series of U[0,1] draws.
my_unif = np.random.uniform(size=30)
pd.DataFrame(my_unif)

# %%
latin_hypercube = dr.getLatinHypercubeDraws(
    sample_size=3, number_of_draws=10, symmetric=False, uniformNumbers=my_unif
)
pd.DataFrame(latin_hypercube)

# %%
# The uniform draws can also be arranged in a two-dimension array
my_unif = dr.getUniform(sample_size=3, number_of_draws=10)
pd.DataFrame(my_unif)

# %%
latin_hypercube = dr.getLatinHypercubeDraws(
    sample_size=3, number_of_draws=10, uniformNumbers=my_unif
)
pd.DataFrame(latin_hypercube)

# %%
# Halton draws
# ------------

# %%
# One Halton sequence.
halton = dr.getHaltonDraws(sample_size=2, number_of_draws=10, base=3)
pd.DataFrame(halton)

# %%
# Several Halton sequences.
halton = dr.getHaltonDraws(sample_size=3, number_of_draws=10)
pd.DataFrame(halton)

# %%
# Shuffled Halton sequences.
halton = dr.getHaltonDraws(sample_size=3, number_of_draws=10, shuffled=True)
pd.DataFrame(halton)

# %%
# The above sequences were generated using the default base: 2. It is
# possible to generate sequences using different prime numbers.
halton = dr.getHaltonDraws(sample_size=1, number_of_draws=10, base=3)
pd.DataFrame(halton)

# %%
# It is also possible to skip the first items of the sequence. This is
# desirable in the context of Monte-Carlo integration.
halton = dr.getHaltonDraws(sample_size=1, number_of_draws=10, base=3, skip=10)
pd.DataFrame(halton)

# %%
# Antithetic draws
# ----------------

# %%
# Antithetic draws can be generated from any function generating uniform draws.
draws = dr.getAntithetic(dr.getUniform, sample_size=3, number_of_draws=10)
pd.DataFrame(draws)

# %%
# Antithetic MLHS
draws = dr.getAntithetic(dr.getLatinHypercubeDraws, sample_size=3, number_of_draws=10)
pd.DataFrame(draws)

# %%
# Antithetic Halton.
draws = dr.getAntithetic(dr.getHaltonDraws, sample_size=1, number_of_draws=10)
pd.DataFrame(draws)


# %%
# As antithetic Halton draws may be correlated, it is a good idea to
# skip the first draws.
def uniform_halton(sample_size: int, number_of_draws: int) -> np.ndarray:
    """Function generating uniform draws for the antithetic draws"""
    return dr.getHaltonDraws(number_of_draws, sample_size, skip=100)


# %%
draws = dr.getAntithetic(uniform_halton, sample_size=3, number_of_draws=10)
pd.DataFrame(draws)

# %%
# Normal draws
# ------------

# %%
# Generate pseudo-random numbers from a normal distribution N(0,1)
# using the Algorithm AS241 Appl. Statist. (1988) Vol. 37, No. 3 by
# Wichura
draws = dr.getNormalWichuraDraws(sample_size=3, number_of_draws=10)
pd.DataFrame(draws)

# %%
# The antithetic version actually generates half of the draws and
# complete them with their antithetic version
draws = dr.getNormalWichuraDraws(sample_size=3, number_of_draws=10, antithetic=True)
pd.DataFrame(draws)

# %%
# The user can provide her own series of U[0,1] draws. In this
# example, we use the MLHS procedure to generate these draws. Note
# that, if the antithetic version is used, only half of the requested
# draws must be provided.
my_unif = dr.getLatinHypercubeDraws(sample_size=3, number_of_draws=5)
pd.DataFrame(my_unif)

# %%
draws = dr.getNormalWichuraDraws(
    sample_size=3, number_of_draws=10, uniformNumbers=my_unif, antithetic=True
)
pd.DataFrame(draws)

# %%
# The same with Halton draws.
my_unif = dr.getHaltonDraws(sample_size=2, number_of_draws=5, base=3, skip=10)
pd.DataFrame(my_unif)

# %%
draws = dr.getNormalWichuraDraws(
    number_of_draws=10, sample_size=2, uniformNumbers=my_unif, antithetic=True
)
pd.DataFrame(draws)



"""Arithmetic expressions accepted by Biogeme: numerical integration

Michel Bierlaire
10.04.2025 09:27
"""

from __future__ import annotations

import logging

import jax
import jax.numpy as jnp
from biogeme.floating_point import JAX_FLOAT
from numpy.polynomial.hermite import hermgauss

from .base_expressions import ExpressionOrNumeric
from .elementary_expressions import (
    Elementary,
    TypeOfElementaryExpression,
)
from .jax_utils import JaxFunctionType
from .random_variable import RandomVariable
from .unary_expressions import UnaryOperator
from ..exceptions import BiogemeError

logger = logging.getLogger(__name__)


class IntegrateNormal(UnaryOperator):
    """
    Numerical integration
    """

    def __init__(
        self,
        child: ExpressionOrNumeric,
        name: str,
        number_of_quadrature_points: int = 30,
    ):
        """Constructor

        :param child: first arithmetic expression
        :type child: biogeme.expressions.Expression
        :param name: name of the random variable for the integration.
        :type name: string
        """
        super().__init__(child)
        self.random_variable_name: str = name
        self.random_variable_id: int | None = (
            None  # Index of the element in its own array.
        )
        self.number_of_quadrature_points: int = number_of_quadrature_points

    def deep_flat_copy(self) -> IntegrateNormal:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_child = self.child.deep_flat_copy()
        return type(self)(
            child=copy_child,
            name=self.random_variable_name,
            number_of_quadrature_points=self.number_of_quadrature_points,
        )

    def dict_of_elementary_expression(
        self, the_type: TypeOfElementaryExpression
    ) -> dict[str:Elementary]:
        """Extract a dict with all elementary expressions of a specific type

        :param the_type: the type of expression
        :type  the_type: TypeOfElementaryExpression
        """
        the_dict = self.child.dict_of_elementary_expression(the_type=the_type)
        if the_type == TypeOfElementaryExpression.RANDOM_VARIABLE:
            return the_dict | {
                self.random_variable_name: RandomVariable(
                    name=self.random_variable_name
                )
            }
        return the_dict

    def set_specific_id(self, name, specific_id, the_type: TypeOfElementaryExpression):
        """The elementary IDs identify the position of each element in the corresponding datab"""
        if name == self.random_variable_name:
            if the_type != TypeOfElementaryExpression.RANDOM_VARIABLE:
                error_msg = f'Elementary expression {name} is not a random variable to be used for integration.'
                raise BiogemeError(error_msg)
            self.random_variable_id = specific_id

        for child in self.get_children():
            child.set_specific_id(name, specific_id, the_type)

    @property
    def safe_rv_id(self) -> int:
        """Check the presence of the random variable ID before its usage"""
        if self.random_variable_id is None:
            raise BiogemeError(
                f"No id defined for random variable {self.random_variable_name} inside integration expression."
            )
        return self.random_variable_id

    def __str__(self) -> str:
        return f'Integrate({self.child}, "{self.random_variable_name}")'

    def __repr__(self) -> str:
        return f'Integrate({self.child}, "{self.random_variable_name}")'

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        child_jax = jax.checkpoint(
            self.child.recursive_construct_jax_function(
                numerically_safe=numerically_safe
            )
        )
        x, w = hermgauss(self.number_of_quadrature_points)
        x = jnp.asarray(x, dtype=JAX_FLOAT)
        w = jnp.asarray(w, dtype=JAX_FLOAT)

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:

            z_vals = jnp.sqrt(2.0) * x

            def integrand(z_val):
                updated_rv = the_random_variables.at[self.safe_rv_id].set(z_val)
                val = child_jax(parameters, one_row, the_draws, updated_rv)
                return val

            values = jax.vmap(integrand)(z_vals)  # shape: (n_points, sample_size)
            result = jnp.sum(values * w) / jnp.sqrt(jnp.pi)
            return result

        return the_jax_function



import unittest

from biogeme.biogeme import BIOGEME
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta, Elem, NormalCdf, log

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)

# As we estimate a binary model, we remove observations where
# Swissmetro was chosen (CHOICE == 2). We also remove observations
# where one of the two alternatives is not available.

exclude = (TRAIN_AV_SP == 0) + (CAR_AV_SP == 0) + (CHOICE == 2) + (
    (PURPOSE != 1) * (PURPOSE != 3)
) > 0


database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)


TRAIN_COST = TRAIN_CO * (GA == 0)


# We estimate a binary probit model. There are only two alternatives.
V1 = B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate choice probability with the numbering of alternatives

P = {1: NormalCdf(V1 - V3), 3: NormalCdf(V3 - V1)}


prob = Elem(P, CHOICE)


class test_21(unittest.TestCase):
    def testEstimation(self):
        biogeme = BIOGEME(
            database,
            log(prob),
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -986.1887862710962, 10)


if __name__ == '__main__':
    unittest.main()



"""Defines the interface for a catalog of expressions that may be
considered in a specification

Michel Bierlaire
13.04.2025 17:02
"""

import abc
import logging
from typing import Iterator, NamedTuple, final

from biogeme import exceptions

from .base_expressions import Expression

logger = logging.getLogger(__name__)

SEPARATOR = ';'
SELECTION_SEPARATOR = ':'


class NamedExpression(NamedTuple):
    name: str
    expression: Expression


class CatalogItem(NamedTuple):
    catalog_name: str
    item_index: int
    item_name: str


def delegate_to_selected(method_name: str):
    """
    Create a method that delegates its call to `self.selected()[1].<method_name>(...)`.

    This is useful for classes that wrap or dispatch to another `Expression` object,
    avoiding code duplication for common method overrides.

    :param method_name:
        The name of the method to delegate. It must be a valid method name on the
        `Expression` returned by `self.selected()[1]`.
    :return:
        A method that can be assigned to a class and will delegate the call.
    """

    def method(self, *args, **kwargs):
        _, expr = self.selected()
        return getattr(expr, method_name)(*args, **kwargs)

    method.__name__ = method_name
    return method


class MultipleExpression(Expression, metaclass=abc.ABCMeta):
    """Interface for catalog of expressions that are interchangeable. Only one of
    them defines the specification. They are designed to be
    modified algorithmically.
    """

    def __init__(self, the_name: str):
        self.name = the_name  # The name of the expression catalog
        if SEPARATOR in name or SELECTION_SEPARATOR in name:
            error_msg = (
                f'Invalid name: {the_name}. Cannot contain characters '
                f'{SELECTION_SEPARATOR} or {SELECTION_SEPARATOR}'
            )
            raise exceptions.BiogemeError(error_msg)
        super().__init__()

    def deep_flat_copy(self) -> Expression:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` such as this one
        is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        the_expression = self.selected_expression()
        return the_expression.deep_flat_copy()

    @abc.abstractmethod
    def selected(self) -> NamedExpression:
        """Return the selected expression and its name

        :return: the name and the selected expression
        :rtype: tuple(str, biogeme.expressions.Expression)
        """

    @abc.abstractmethod
    def get_iterator(self) -> Iterator[NamedExpression]:
        """Returns an iterator on NamedExpression"""

    @final
    def catalog_size(self) -> int:
        """Provide the size of the catalog

        :return: number of expressions in the catalog
        :rtype: int
        """
        the_iterator = self.get_iterator()

        return len(list(the_iterator))

    @final
    def selected_name(self) -> str:
        """Obtain the name of the selection

        :return: the name of the selected expression
        :rtype: str
        """
        the_name, _ = self.selected()
        return the_name

    @final
    def selected_expression(self) -> NamedExpression:
        """Obtain the selected expression

        :return: the selected expression
        :rtype: biogeme.expressions.Expression
        """
        _, the_expression = self.selected()
        return the_expression

    def __str__(self) -> str:
        named_expression: NamedExpression = self.selected()
        return f'[{self.name}: {named_expression.name}]{named_expression.expression}'


# We now specify all methods that must be delegated to the selected expression
delegated_methods = [
    'set_maximum_number_of_observations_per_individual',
    'change_init_values',
    'rename_elementary',
    'dict_of_elementary_expression',
    'set_of_elementary_expression',
    'get_elementary_expression',
    'get_value',
    'requires_draws',
    'recursive_construct_jax_function',
    'dict_of_draw_types',
    'logit_choice_avail',
    'add_suffix_to_all_variables',
    'set_specific_id',
]

for name in delegated_methods:
    setattr(MultipleExpression, name, final(delegate_to_selected(name)))



"""Functions to validate the arithmetic expressions

Michel Bierlaire
Tue Mar 25 17:17:11 2025
"""

from typing import Any


from .numeric_tools import is_numeric
from ..exceptions import BiogemeError


def validate_expression_type(expression: Any) -> None:
    """
    :param expression: expression to validate
    """
    from .base_expressions import Expression

    if not (is_numeric(expression) or isinstance(expression, Expression)):
        error_msg = f'Invalid expression: {str(expression)}'
        raise BiogemeError(error_msg)



""" Defines a class that represents a partition of a list of integers

:author: Michel Bierlaire
:date: Thu Oct 26 17:30:36 2023

"""

import logging
from typing import Iterator, Sequence

Segment = set[int]

logger = logging.getLogger(__name__)


class Partition:
    """Class defining a partition of a set"""

    def __init__(self, segments: Sequence[Segment], full_set: Segment | None = None):
        self.segments = segments
        if full_set:
            self.full_set = full_set
        else:
            logger.warning(
                'The full set of the partition has not been specified. It is defined '
                'by default as the union of the provided segment.'
            )
            self.full_set = set().union(*self.segments)
        self.validate_segments()
        self.validate_partition()

    def number_of_segments(self) -> int:
        return len(self.segments)

    def validate_segments(self) -> None:
        """Check the types"""
        for segment in self.segments:
            if not isinstance(segment, set):
                raise ValueError('Each segment must be of type "set".')
            if not segment:
                raise ValueError('Partition contains empty segment')
            for item in segment:
                if not isinstance(item, int):
                    raise ValueError('Each item in the segment must be an integer.')

    def validate_partition(self) -> None:
        """Check that there is no intersection between segments"""
        for s1, segment1 in enumerate(self.segments):
            for s2, segment2 in enumerate(self.segments):
                if s1 != s2:
                    intersection = segment1.intersection(segment2)
                    if intersection:
                        raise ValueError(
                            f'Elements in two different segments: {intersection}.'
                        )

        # Check that the union of all segments is the given set
        union_of_segments = set().union(*self.segments)
        if union_of_segments != self.full_set:
            missing_elements = self.full_set - union_of_segments
            raise ValueError(
                f'The union of segments does not match the given set. '
                f'Missing elements: {missing_elements}.'
            )

    def __iter__(self) -> Iterator[Segment]:
        return iter(self.segments)



"""Arithmetic expressions accepted by Biogeme: Elem

Michel Bierlaire
Fri Apr 25 2025, 10:33:58
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
from biogeme.exceptions import BiogemeError

from .base_expressions import Expression
from .beta_parameters import Beta
from .convert import validate_and_convert
from .jax_utils import JaxFunctionType

if TYPE_CHECKING:
    from . import ExpressionOrNumeric

logger = logging.getLogger(__name__)


class Elem(Expression):
    """This returns the element of a dictionary. The key is evaluated
    from an expression and must return an integer, possibly negative.
    """

    def __init__(
        self,
        dict_of_expressions: dict[int, ExpressionOrNumeric],
        key_expression: ExpressionOrNumeric,
    ):
        """Constructor

        :param dict_of_expressions: dict of objects with numerical keys.
        :type dict_of_expressions: dict(int: biogeme.expressions.Expression)

        :param key_expression: object providing the key of the element
                              to be evaluated.
        :type key_expression: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        super().__init__()

        self.key_expression = validate_and_convert(key_expression)
        self._key_depends_on_parameters = self.key_expression.embed_expression(Beta)
        self.children.append(self.key_expression)

        self.dict_of_expressions = {}  #: dict of expressions
        for k, v in dict_of_expressions.items():
            self.dict_of_expressions[k] = validate_and_convert(v)
            self.children.append(self.dict_of_expressions[k])

    def deep_flat_copy(self) -> Elem:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_dict_of_expressions = {
            key: expression.deep_flat_copy()
            for key, expression in self.dict_of_expressions.items()
        }
        copy_key = self.key_expression.deep_flat_copy()
        return type(self)(
            dict_of_expressions=copy_dict_of_expressions, key_expression=copy_key
        )

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float

        :raise BiogemeError: if the calculated key is not present in
            the dictionary.
        """
        try:
            key = int(self.key_expression.get_value())
            return self.dict_of_expressions[key].get_value()
        except (ValueError, KeyError):
            raise BiogemeError(
                f'Invalid or missing key: {key}. '
                f'Available keys: {self.dict_of_expressions.keys()}'
            )

    def __str__(self) -> str:
        s = '{{'
        first = True
        for k, v in self.dict_of_expressions.items():
            if first:
                s += f'{k}:{v}'
                first = False
            else:
                s += f', {k}:{v}'
        s += f'}}[{self.key_expression}]'
        return s

    def __repr__(self) -> str:
        return f"Elem({repr(self.dict_of_expressions)}, {repr(self.key_expression)})"

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        compiled_dict = {
            k: v.recursive_construct_jax_function(numerically_safe=numerically_safe)
            for k, v in self.dict_of_expressions.items()
        }
        key_fn = self.key_expression.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        sorted_keys = sorted(compiled_dict)
        key_array = jnp.array(sorted_keys)

        def make_branch(fn, k):
            def wrapped(*args):

                result = fn(*args)
                return result

            return wrapped

        branches = [make_branch(compiled_dict[k], k) for k in sorted_keys]

        # branches = [
        #    lambda p, r, d, rv, fn=compiled_dict[k]: fn(p, r, d, rv)
        #    for k in sorted_keys
        # ]

        def the_jax_function(parameters, one_row, the_draws, the_random_variables):
            key_value = key_fn(parameters, one_row, the_draws, the_random_variables)
            key_int = jnp.asarray(key_value, dtype=jnp.int32)
            matches = key_array == key_int
            branch_index = jnp.argmax(matches)
            result = jax.lax.switch(
                branch_index,
                branches,
                parameters,
                one_row,
                the_draws,
                the_random_variables,
            )
            return result

        return the_jax_function



"""

One model among many
====================

We consider the model with 432 specifications defined in
:ref:`everything_spec_section`. We select one specification and estimate it.
See `Bierlaire and Ortelli (2023)
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_.

:author: Michel Bierlaire, EPFL
:date: Sat Jul 15 15:46:56 2023

"""
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from everything_spec import model_catalog, database, av

logger = blog.get_screen_logger(level=blog.INFO)

# %%
# The code characterizing the specification should be copied from the
# .pareto file generated by the algorithm, or from one of the
# glossaries illustrated in earlier examples.
SPEC_ID = (
    'ASC:GA-LUGGAGE;'
    'B_COST_gen_altspec:generic;'
    'B_TIME:FIRST;'
    'B_TIME_gen_altspec:generic;'
    'model_catalog:logit;'
    'train_tt_catalog:power'
)

# %% The biogeme object for the selected model can be obtained from
# the spec_id, and used as usual.
the_biogeme = bio.BIOGEME.from_configuration(
    config_id=SPEC_ID,
    expression=model_catalog,
    database=database,
)
the_biogeme.modelName = 'my_favorite_model'

# %%
# Calculate of the null log-likelihood for reporting.
the_biogeme.calculateNullLoglikelihood(av)

# %%
# Estimate the parameters.
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
# Get the results in a pandas table
pandas_results = results.getEstimatedParameters()

# %%
pandas_results
